# -*- coding: utf-8 -*-

#Copyright (c) 2017 André Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

###############################################################################
# Imports
###############################################################################

from builtins import range
from collections import Counter, namedtuple
from itertools import chain
import os

import magic as file_cmd

###############################################################################
# Notes
###############################################################################

# It seems that it is better to store string names for source objects, instead
# of RosName. If necessary, store namespace also as a string. Source objects
# will be used as generators for runtime objects, so we must have access to
# the values provided originally.

# Source objects define various types of dependencies (files, packages, etc.).
# The objects exist, but only work properly if the dependencies are met.
# Runtime objects do not really have dependencies (only the node instances
# depend on the node executable). Instead, they have conditions, boolean
# expressions that determine whether the objects will be instantiated.


###############################################################################
# Base Metamodel Object
###############################################################################

class MetamodelObject(object):
    pass


###############################################################################
# Analysis Properties
###############################################################################

class DependencySet(object):
    def __init__(self):
        self.files          = set()
        self.packages       = set()
        self.arguments      = set()
        self.environment    = set()

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return (self.files == other.files
                and self.packages == other.packages
                and self.arguments == other.arguments
                and self.environment == other.environment)

    def __ne__(self, other):
        return not self.__eq__(other)


class SourceCondition(object):
    def __init__(self, condition, location=None, statement="if"):
        self.statement = statement
        self.condition = condition
        self.location = location
        self.location2 = loc_to_loc2(location)

    @property
    def language(self):
        if self.location and self.location.file:
            return self.location.file.language
        return "unknown"

    def to_JSON_object(self):
        return {
            "statement": self.statement,
            "condition": str(self.condition),
            "location": loc2_to_JSON(self.location2)
        }

    def __str__(self):
        return "{} {}".format(self.statement, str(self.condition))

    def __repr__(self):
        return self.__str__()


class _UnknownValue(object):
    __slots__ = ("name", "_call", "_attr", "_fun")

    def __init__(self, name, call, attr, fun):
        self.name = name
        self._call = call
        self._attr = attr
        self._fun = fun

    @property
    def attr(self):
        return self._attr

    def resolve(self, value):
        if value is not None:
            setattr(self._call, self._attr, self._fun(value))
            return self.name in self._call._vars
        return False

    def __str__(self):
        return self.name


class RosPrimitiveCall(MetamodelObject):
    KEY = ""

    """"Base class for calls to ROS primitives."""
    def __init__(self, name, namespace, msg_type, control_depth = None,
                 repeats = False, conditions = None, location = None):
        self._vars = {}
        self.name = name
        self.namespace = namespace
        self.type = msg_type
        self.conditions = conditions if not conditions is None else []
        self.control_depth = control_depth or len(self.conditions)
        self.repeats = repeats and self.control_depth >= 1
        self.location = location
        self.location2 = loc_to_loc2(location)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._set_str_attr("name", value)

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, value):
        self._set_str_attr("namespace", value)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._set_str_attr("type", value)

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value is None or value.package is None or value.file is None:
            if "location" not in self._vars:
                self._vars["location"] = self._new_var("location", fpid)
        else:
            if "location" in self._vars:
                del self._vars["location"]
        self._location = value
        self.location2 = loc_to_loc2(value)

    @property
    def conditions(self):
        return self._conditions

    @conditions.setter
    def conditions(self, value):
        if value:
            if "conditions" not in self._vars:
                self._vars["conditions"] = self._new_var(
                    "conditions", _bool_to_conditions)
        else:
            if "conditions" in self._vars:
                del self._vars["conditions"]
        self._conditions = value

    @property
    def full_name(self):
        if not self.namespace or self.name.startswith("/"):
            return self.name
        if self.namespace.endswith("/"):
            return self.namespace + self.name
        return self.namespace + "/" + self.name

    @property
    def rostype(self):
        return self.type

    def variables(self):
        return list(self._vars.values())

    def refine_from_JSON_specs(self, data):
        if "?" in self.name:
            self.name = data["name"]
        if "?" in self.namespace:
            self.namespace = data.get("ns", self.namespace)
        self.conditions = []
        self.control_depth = 0
        self.repeats = False
        self.location = data.get("location", self.location)

    def to_JSON_object(self):
        return {
            "name": self.name,
            "namespace": self.namespace,
            "type": self.type,
            "depth": self.control_depth,
            "repeats": self.repeats,
            "conditions": [c.to_JSON_object() for c in self.conditions],
            "location": loc2_to_JSON(self.location2),
            "variables": {attr: uv.name for attr, uv in self._vars.items()}
        }

    def _set_str_attr(self, attr, value):
        if value is None or value == "?":
            if attr not in self._vars:
                self._vars[attr] = self._new_var(attr, str)
        else:
            if attr in self._vars:
                del self._vars[attr]
        setattr(self, "_" + attr, value)

    _var_counter = 0

    def _new_var(self, attr, tf):
        type(self)._var_counter += 1
        name = "{}@{}{}".format(self.KEY, attr, type(self)._var_counter)
        return _UnknownValue(name, self, attr, tf)

    def __str__(self):
        return "RosPrimitiveCall({}, {}, {}) {} (depth {})".format(
            self.name, self.namespace, self.type,
            self.location2, self.control_depth
        )

    def __repr__(self):
        return self.__str__()

class AdvertiseCall(RosPrimitiveCall):
    KEY = "advertise"
    _var_counter = 0

    def __init__(self, name, namespace, msg_type, queue_size, latched=False,
                 control_depth=None, repeats=False, conditions=None,
                 location=None):
        RosPrimitiveCall.__init__(self, name, namespace, msg_type,
                                  control_depth = control_depth,
                                  repeats = repeats,
                                  conditions = conditions, location = location)
        self.queue_size = queue_size
        self.latched = latched

    @property
    def queue_size(self):
        return self._queue_size

    @queue_size.setter
    def queue_size(self, value):
        key = "queue_size"
        if value is None:
            if key not in self._vars:
                self._vars[key] = self._new_var(key, int)
        else:
            if key in self._vars:
                del self._vars[key]
        self._queue_size = value

    @property
    def latched(self):
        return self._latched

    @latched.setter
    def latched(self, value):
        key = "latched"
        if value is None:
            if key not in self._vars:
                self._vars[key] = self._new_var(key, bool)
        else:
            if key in self._vars:
                del self._vars[key]
        self._latched = value

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        msg_type = datum["msg_type"]
        queue = datum["queue_size"]
        latched = datum.get("latched", False)
        return cls(name, ns, msg_type, queue, latched=latched)

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["msg_type"]
        self.queue_size = data.get("queue_size", self.queue_size)
        self.latched = data.get("latched", self.latched)

    def to_JSON_object(self):
        data = RosPrimitiveCall.to_JSON_object(self)
        data["queue"] = self.queue_size
        data["latched"] = self.latched
        return data

    def clone(self):
        call = AdvertiseCall(self.name, self.namespace, self.type,
            self.queue_size, latched=self.latched,
            control_depth=self.control_depth, repeats=self.repeats,
            conditions=list(self.conditions), location=self.location)
        call.location2 = self.location2
        return call

Publication = AdvertiseCall

class SubscribeCall(RosPrimitiveCall):
    KEY = "subscribe"
    _var_counter = 0

    def __init__(self, name, namespace, msg_type, queue_size,
                 control_depth = None, repeats = False, conditions = None,
                 location = None):
        RosPrimitiveCall.__init__(self, name, namespace, msg_type,
                                  control_depth = control_depth,
                                  repeats = repeats,
                                  conditions = conditions, location = location)
        self.queue_size = queue_size

    @property
    def queue_size(self):
        return self._queue_size

    @queue_size.setter
    def queue_size(self, value):
        key = "queue_size"
        if value is None:
            if key not in self._vars:
                self._vars[key] = self._new_var(key, int)
        else:
            if key in self._vars:
                del self._vars[key]
        self._queue_size = value

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        msg_type = datum["msg_type"]
        queue = datum["queue_size"]
        return cls(name, ns, msg_type, queue)

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["msg_type"]
        self.queue_size = data.get("queue_size", self.queue_size)

    def to_JSON_object(self):
        data = RosPrimitiveCall.to_JSON_object(self)
        data["queue"] = self.queue_size
        return data

    def clone(self):
        call = SubscribeCall(self.name, self.namespace, self.type,
            self.queue_size, control_depth=self.control_depth,
            repeats=self.repeats, conditions=list(self.conditions),
            location=self.location)
        call.location2 = self.location2
        return call

Subscription = SubscribeCall

class AdvertiseServiceCall(RosPrimitiveCall):
    KEY = "advertiseService"
    _var_counter = 0

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["srv_type"]

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        srv_type = datum["srv_type"]
        return cls(name, ns, srv_type)

    def clone(self):
        call = AdvertiseServiceCall(self.name, self.namespace, self.type,
            control_depth=self.control_depth, repeats=self.repeats,
            conditions=list(self.conditions), location=self.location)
        call.location2 = self.location2
        return call

ServiceServerCall = AdvertiseServiceCall

class ServiceClientCall(RosPrimitiveCall):
    KEY = "serviceClient"
    _var_counter = 0

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["srv_type"]

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        srv_type = datum["srv_type"]
        return cls(name, ns, srv_type)

    def clone(self):
        call = ServiceClientCall(self.name, self.namespace, self.type,
            control_depth=self.control_depth, repeats=self.repeats,
            conditions=list(self.conditions), location=self.location)
        call.location2 = self.location2
        return call

class GetParamCall(RosPrimitiveCall):
    KEY = "getParam"
    _var_counter = 0

    def __init__(self, name, namespace, param_type, default_value=None,
                 control_depth=None, repeats=False, conditions=None,
                 location=None):
        RosPrimitiveCall.__init__(self, name, namespace, param_type,
            control_depth=control_depth, repeats=repeats,
            conditions=conditions, location=location)
        self.default_value = default_value

    @property
    def default_value(self):
        return self._default_value

    @default_value.setter
    def default_value(self, value):
        key = "default_value"
        if value is None:
            if key not in self._vars:
                self._vars[key] = self._new_var(key, fpid)
        else:
            if key in self._vars:
                del self._vars[key]
        self._default_value = value

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        param_type = datum["param_type"]
        v = datum.get("default_value")
        return cls(name, ns, param_type, default_value=v)

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["param_type"]
        self.default_value = data.get("default_value", self.default_value)

    def to_JSON_object(self):
        data = RosPrimitiveCall.to_JSON_object(self)
        data["default_value"] = self.default_value
        return data

    def clone(self):
        call = GetParamCall(self.name, self.namespace, self.type,
            default_value=self.default_value,
            control_depth=self.control_depth, repeats=self.repeats,
            conditions=list(self.conditions), location=self.location)
        call.location2 = self.location2
        return call

ReadParameterCall = GetParamCall

class SetParamCall(RosPrimitiveCall):
    KEY = "setParam"
    _var_counter = 0

    def __init__(self, name, namespace, param_type, value=None,
                 control_depth=None, repeats=False, conditions=None,
                 location=None):
        RosPrimitiveCall.__init__(self, name, namespace, param_type,
            control_depth=control_depth, repeats=repeats,
            conditions=conditions, location=location)
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        key = "value"
        if value is None:
            if key not in self._vars:
                self._vars[key] = self._new_var(key, fpid)
        else:
            if key in self._vars:
                del self._vars[key]
        self._value = value

    @classmethod
    def from_JSON_specs(cls, datum):
        name = datum["name"]
        ns = datum.get("namespace", "")
        param_type = datum["param_type"]
        v = datum.get("value")
        return cls(name, ns, param_type, value=v)

    def refine_from_JSON_specs(self, data):
        RosPrimitiveCall.refine_from_JSON_specs(self, data)
        self.type = data["param_type"]
        self.value = data.get("value", self.value)

    def to_JSON_object(self):
        data = RosPrimitiveCall.to_JSON_object(self)
        data["value"] = self.value
        return data

    def clone(self):
        call = SetParamCall(self.name, self.namespace, self.type,
            value=self.value,
            control_depth=self.control_depth, repeats=self.repeats,
            conditions=list(self.conditions), location=self.location)
        call.location2 = self.location2
        return call

WriteParameterCall = SetParamCall


###############################################################################
# Source Code Structures
###############################################################################

class Person(object):
    """Represents a person (author/maintainer)."""
    def __init__(self, name, email = "email@example.com"):
        self.id = email
        self.name = name or "Unknown"
        self.email = email

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.id.__hash__()


Location2 = namedtuple("Location2", ("package", "file", "line", "column"))

# in the process of deprecation
class Location(object):
    """A location to report (package, file, line)."""
    def __init__(self, pkg, file=None, line=None, col=None, fun=None, cls=None):
        self.package = pkg
        self.file = file
        self.line = line    # should start at 1
        self.column = col   # should start at 1
        self.function = fun
        self.class_ = cls

    @property
    def largest_scope(self):
        return self.package

    @property
    def smallest_scope(self):
        return self.file or self.package

    def to_JSON_object(self):
        return {
            "package": self.package.name,
            "file": self.file.full_name if self.file else None,
            "line": self.line,
            "column": self.column,
            "function": self.function,
            "class": self.class_
        }

    def __str__(self):
        s = "in " + self.package.name
        if not self.file:
            return s
        s += "/" + self.file.full_name
        if not self.line is None:
            s += ":" + str(self.line)
            if not self.column is None:
                s += ":" + str(self.column)
            if self.function:
                s += ", in function " + self.function
            if self.class_:
                s += ", in class " + self.class_
        return s


class SourceObject(MetamodelObject):
    """Base class for objects subject to analysis."""
    SCOPES = ("file", "node", "package", "repository", "project")

    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.dependencies = DependencySet()
        self._analyse = True

    @property
    def location(self):
        return None

    def __lt__(self, scope):
        if isinstance(scope, basestring):
            return self.SCOPES.index(self.scope) < self.SCOPES.index(scope)
        return self.SCOPES.index(self.scope) < self.SCOPES.index(scope.scope)

    def __le__(self, scope):
        if isinstance(scope, basestring):
            return self.SCOPES.index(self.scope) <= self.SCOPES.index(scope)
        return self.SCOPES.index(self.scope) <= self.SCOPES.index(scope.scope)

    def __gt__(self, scope):
        if isinstance(scope, basestring):
            return self.SCOPES.index(self.scope) > self.SCOPES.index(scope)
        return self.SCOPES.index(self.scope) > self.SCOPES.index(scope.scope)

    def __ge__(self, scope):
        if isinstance(scope, basestring):
            return self.SCOPES.index(self.scope) >= self.SCOPES.index(scope)
        return self.SCOPES.index(self.scope) >= self.SCOPES.index(scope.scope)

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.id.__hash__()

    def bound_to(self, scope):
        return self == scope

    def accepts_scope(self, scope):
        return self.scope == scope

    def __str__(self):
        return self.id


class SourceFile(SourceObject):
    """Represents a source code file."""
    CPP = ('c source', 'c++ source')
    CPP_ALT = ('.c', '.h', '.hpp', '.hxx', '.cpp', '.cc', '.cxx')
    PYTHON = 'python script'
    PYTHON_ALT = '.py'
    PKG_XML = 'package.xml'
    LAUNCH = ('.launch', '.launch.xml')
    MSG = '.msg'
    SRV = '.srv'
    ACTION = '.action'
    YAML = ('.yaml', '.yml')
    CMAKELISTS = 'CMakeLists.txt'

    def __init__(self, name, directory, pkg):
        id = ("file:" + pkg.name + "/" + directory.replace(os.path.sep, "/")
              + "/" + name)
        SourceObject.__init__(self, id, name)
        self.directory = directory
        self.full_name = os.path.join(directory, name)
        self.dir_path = os.path.join(pkg.path, directory)
        self.path = os.path.join(pkg.path, directory, name)
        self.package = pkg
        self.language = self._get_language()
        self.tree = None
        self.size = 0
        self.lines = 0
        self.sloc = 0
        self.timestamp = 0

    @property
    def scope(self):
        return "file"

    @property
    def location(self):
        return Location(self.package, file=self)

    def bound_to(self, other):
        if other.scope == "node":
            return self in other.source_files
        if other.scope == "package":
            return self.package == other
        if other.scope == "repository" or other.scope == "project":
            return self.package in other.packages
        return other.scope == "file" and self == other

    def accepts_scope(self, scope):
        return self >= scope

    def set_file_stats(self):
        self.size = os.path.getsize(self.path)
        self.timestamp = os.path.getmtime(self.path)
        self.lines = 0
        self.sloc = 0
        ignore_all = []
        to_ignore = {"*": ignore_all}
        ilp, inlp = self._ignore_parsers()
        with open(self.path, "r") as handle:
            for line in handle:
                self.lines += 1
                sline = line.strip()
                if sline:
                    self.sloc += 1
                    if ilp(sline):
                        ignore_all.append(self.lines)
                    elif inlp(sline):
                        ignore_all.append(self.lines + 1)
        return to_ignore

    def to_JSON_object(self):
        return {
            "name": self.name,
            "directory": self.directory,
            "package": self.package.name,
            "language": self.language,
            "size": self.size,
            "timestamp": self.timestamp,
            "lines": self.lines,
            "sloc": self.sloc
        }

    def _get_language(self):
        try:
            file_type = file_cmd.from_file(self.path).lower()
            if file_type.startswith(self.CPP):
                return 'cpp'
            if self.PYTHON in file_type:
                return 'python'
        except IOError:
            if self.name.endswith(self.CPP_ALT):
                return 'cpp'
            if self.name.endswith(self.PYTHON_ALT):
                return 'python'
        if self.name.endswith(self.LAUNCH):
            return 'launch'
        if self.name == self.PKG_XML:
            return 'package'
        if self.name.endswith(self.MSG):
            return 'msg'
        if self.name.endswith(self.SRV):
            return 'srv'
        if self.name.endswith(self.ACTION):
            return 'action'
        if self.name.endswith(self.YAML):
            return 'yaml'
        if self.name == self.CMAKELISTS:
            return 'cmake'
        return 'unknown'

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.id

    def _ignore_parsers(self):
        if self.language == "cpp":
            return (_cpp_ignore_line, _cpp_ignore_next_line)
        elif self.language == "python":
            return (_py_ignore_line, _py_ignore_next_line)
        return (_no_parser, _no_parser)


class Package(SourceObject):
    """Represents a ROS package."""
    def __init__(self, name, repo = None, proj = None):
        SourceObject.__init__(self, "package:" + name, name)
    # public:
        self.project            = proj
        self.repository         = repo
        self.authors            = set()
        self.maintainers        = set()
        self.is_metapackage     = False
        self.description        = ""
        self.version            = "0.0.0"
        self.licenses           = set()
        self.website            = None
        self.vcs_url            = None
        self.bug_url            = None
        self.path               = None
        self.source_files       = []
        self.nodes              = []
        self.size               = 0 # sum of file sizes
        self.lines              = 0 # sum of physical file lines
        self.sloc               = 0 # sum of file source lines of code
        self.topological_tier   = 0

    @property
    def scope(self):
        return "package"

    @property
    def location(self):
        return Location(self)

    @property
    def file_count(self):
        return len(self.source_files)

    def bound_to(self, other):
        if other.scope == "file" or other.scope == "node":
            return other.package == self
        if other.scope == "repository":
            return self.repository == other
        if other.scope == "project":
            return self.project == other
        if other.scope == "package":
            if self == other:
                return True
            for dep in self.dependencies.packages:
                if dep.type == "package" and dep.value == other.name:
                    return True
        return False

    def to_JSON_object(self):
        return {
            "name": self.name,
            "metapackage": self.is_metapackage,
            "description": self.description,
            "version": self.version,
            "wiki": self.website,
            "repository": self.vcs_url,
            "bugTracker": self.bug_url,
            "authors": [person.name for person in self.authors],
            "maintainers": [person.name for person in self.maintainers],
            "dependencies": [pkg for pkg in self.dependencies.packages],
            "files": [f.full_name for f in self.files],
            "nodes": [n.node_name for n in self.nodes]
        }

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.id


class Repository(SourceObject):
    """Represents a source code repository."""
    def __init__(self, name, vcs = None, url = None, version = None,
                 status = None, path = None, proj = None):
        SourceObject.__init__(self, "repository:" + name, name)
        self.project        = None
        self.vcs            = vcs
        self.url            = url
        self.version        = version
        self.status         = status
        self.path           = path
        self.packages       = []
        self.declared_packages = []
        self.commits        = 1
        self.contributors   = 1

    @property
    def scope(self):
        return "repository"

    def bound_to(self, other):
        if other.scope == "package":
            return other.repository == self
        if other.scope == "file" or other.scope == "node":
            return other.package in self.packages
        if other.scope == "project":
            if self.project == other:
                return True
            for package in self.packages:
                if not package in other.packages:
                    return False
                return True
        return other.scope == "repository" and self == other

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.id


class Project(SourceObject):
    """A project is a custom grouping of packages, not necessarily
        corresponding to a repository, and not even requiring the
        existence of one.
    """
    def __init__(self, name):
        if name == "all":
            raise ValueError("Forbidden project name: all")
        SourceObject.__init__(self, "project:" + name, name)
        self.packages = []
        self.repositories = []
        self.configurations = []
        self.node_specs = {}

    @property
    def scope(self):
        return "project"

    def get_node(self, node_name):
        parts = node_name.split("/")
        if not len(parts) == 2:
            raise ValueError("Expected '<pkg>/<node>' string, got {}.".format(
                repr(node_name)))
        for pkg in self.packages:
            if pkg.name == parts[0]:
                break
        else:
            raise ValueError("Package is not part of the project: " + parts[0])
        for node in pkg.nodes:
            if node.name == parts[1]:
                break
        else:
            raise ValueError("Unknown node: " + node_name)
        return node

    def bound_to(self, other):
        if other.scope == "package":
            return other.project == self
        if other.scope == "file" or other.scope == "node":
            return other.package in self.packages
        if other.scope == "repository":
            if other.project == self:
                return True
            for package in other.packages:
                if not package in self.packages:
                    return False
                return True
        return other.scope == "project" and self == other

    def to_JSON_object(self):
        return {
            "id": self.name,
            "packages": [pkg.name for pkg in self.packages],
            "repositories": [repo.name for repo in self.repositories]
        }

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.id


class Node(SourceObject):
    def __init__(self, name, pkg, rosname = None, nodelet = None):
        id = "node:" + pkg.name + "/" + (nodelet or name)
        SourceObject.__init__(self, id, name)
        self.package = pkg
        if isinstance(rosname, basestring):
            rosname = RosName(rosname)
        self.rosname = rosname
        self.nodelet_class = nodelet
        self.source_files = []
        self.source_tree = None
        self.instances = []
        self.advertise = []
        self.subscribe = []
        self.service = []
        self.client = []
        self.read_param = []
        self.write_param = []
        self.hpl_properties = [] # [string | HplAstObject]
        self.hpl_assumptions = [] # [string | HplAstObject]

    @property
    def scope(self):
        return "node"

    @property
    def location(self):
        return Location(self.package)

    @property
    def is_nodelet(self):
        return not self.nodelet_class is None

    @property
    def language(self):
        for sf in self.source_files:
            return sf.language
        return None

    @property
    def node_name(self):
        return self.package.name + "/" + (self.nodelet_class or self.name)

    @property
    def timestamp(self):
        return max([f.timestamp for f in self.source_files] or [0])

    def variables(self):
        vs = []
        for call in chain(self.advertise, self.subscribe, self.service,
                          self.client, self.read_param, self.write_param):
            vs.extend(v for v in call.variables())
        return vs

    def resolve_variables(self, data):
        for v in self.variables():
            v.resolve(data.get(v.name))

    def to_JSON_object(self):
        return {
            "id": self.node_name,
            "name": self.name,
            "package": self.package.name,
            "rosname": self.rosname,
            "nodelet": self.nodelet_class,
            "files": [f.full_name for f in self.source_files],
            "advertise": [p.to_JSON_object() for p in self.advertise],
            "subscribe": [p.to_JSON_object() for p in self.subscribe],
            "service": [p.to_JSON_object() for p in self.service],
            "client": [p.to_JSON_object() for p in self.client],
            "readParam": [p.to_JSON_object() for p in self.read_param],
            "writeParam": [p.to_JSON_object() for p in self.write_param],
            "timestamp": self.timestamp,
            "hpl": {
                "properties": map(str, self.hpl_properties),
                "assumptions": map(str, self.hpl_assumptions)
            }
        }

    def bound_to(self, other):
        if other.scope == "package":
            return other == self.package
        if other.scope == "file":
            return other in self.source_files
        if other.scope == "repository" or other.scope == "project":
            for package in other.packages:
                if package == self.package:
                    return True
                return False
        return other.scope == "node" and self == other

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.id


###############################################################################
# ROS Computation Graph
###############################################################################

class RosName(object):
    def __init__(self, name, ns = "/", private_ns = "", remaps = None):
        self._given = name
        self._name = RosName.transform(name, ns = ns, private_ns = private_ns,
                                       remaps = remaps)
        parts = self._name.rsplit("/", 1)
        self._own = parts[-1]
        self._ns = parts[0] or "/"

    @property
    def full(self):
        return self._name

    @property
    def own(self):
        return self._own

    @property
    def namespace(self):
        return self._ns

    @property
    def given(self):
        return self._given

    @property
    def is_global(self):
        return self._given.startswith("/")

    @property
    def is_private(self):
        return self._given.startswith("~")

    @property
    def is_unresolved(self):
        return "?" in self._name

    @property
    def pattern(self):
        return RosName.to_pattern(self._name)

    @staticmethod
    def to_pattern(name):
        parts = []
        prev = ""
        n = len(name)
        i = 0
        if name == "?":
            return ".+$"
        if name[0] == "?":
            parts.append("(.*?)")
            i = 1
            prev = "?"
            assert name[1] != "?"
        for j in xrange(i, n):
            if name[j] == "?":
                assert prev != "?"
                if prev == "/":
                    if j == n - 1: # self._name.endswith("/?")
                        # end, whole part for sure
                        parts.append(name[i:j])
                        parts.append("(.+?)")
                    elif name[j+1] == "/": # "/?/"
                        # start and middle, whole parts
                        parts.append(name[i:j-1])
                        parts.append("(/.+?)?")
                    else: # "/?a", optional part
                        parts.append(name[i:j])
                        parts.append("(.*?)")
                else: # "a?/", "a?a", "/a?", optional part
                    parts.append(name[i:j])
                    parts.append("(.*?)")
                i = j + 1
            prev = name[j]
        if i < n:
            parts.append(name[i:])
        parts.append("$")
        return "".join(parts)

    @staticmethod
    def resolve(name, ns="/", private_ns=""):
        if name.startswith("~"):
            return private_ns + "/" + name[1:]
        elif name.startswith("/"):
            return name
        elif ns.endswith("/"):
            return ns + name
        else:
            return ns + "/" + name

    @staticmethod
    def transform(name, ns="/", private_ns="", remaps=None):
        name = RosName.resolve(name, ns=ns, private_ns=private_ns)
        if remaps:
            return remaps.get(name, name)
        return name

    @staticmethod
    def resolve_ns(new_ns, ns="/", private_ns=""):
        if not new_ns:
            return ns
        if new_ns.startswith("~"):
            if len(new_ns) == 1:
                return private_ns
            return "{}/{}".format(private_ns, new_ns[1:])
        return RosName.resolve(new_ns, ns=ns, private_ns=private_ns)

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self._name == other._name
        return self._name == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self._name.__hash__()

    def __str__(self):
        return self._name

    def __repr__(self):
        return "{}({!r}, ns={!r})".format(
            type(self).__name__, self._own, self._ns)


class RuntimeLocation(object):
    def __init__(self, configuration):
        self.configuration = configuration

    @property
    def largest_scope(self):
        return self.configuration

    @property
    def smallest_scope(self):
        return self.configuration

    def to_JSON_object(self):
        return {"configuration": self.configuration.name}

    def __str__(self):
        return "in configuration " + self.configuration.name


class Resource(MetamodelObject):
    """This is the base class for all runtime objects belonging
        to the ROS Computation Graph.
    """

    def __init__(self, config, rosname, conditions = None):
        self.configuration = config
        self.rosname = rosname
        self.conditions = conditions if not conditions is None else []
        self.variables = []

    @property
    def id(self):
        return self.rosname.full

    @property
    def name(self):
        return self.rosname.own

    @property
    def namespace(self):
        return self.rosname.namespace

    @property
    def enabled(self):
        return not self.conditions

    @property
    def disabled(self):
        return False in self.conditions

    @property
    def conditional(self):
        return not self.enabled and not self.disabled

    @property
    def unresolved(self):
        return self.rosname.is_unresolved

    @property
    def location(self):
        return RuntimeLocation(self.configuration)

    @property
    def resource_type(self):
        raise NotImplementedError("subclasses must implement this property")

    def traceability(self):
        raise NotImplementedError("subclasses must implement this method")

    def remap(self, rosname):
        raise NotImplementedError("subclasses must implement this method")

    def resolve_name(self, name):
        if not name:
            return self.namespace
        return RosName.resolve(name, ns=self.namespace, private_ns=self.id)

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return (self.configuration == other.configuration
                    and self.rosname.full == other.rosname.full)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.rosname.__hash__()

    def __str__(self):
        return "{}('{}', {})".format(self.__class__.__name__,
            self.rosname.full, self.type)

    def __repr__(self):
        return self.__str__()


class NodeInstance(Resource):
    def __init__(self, config, rosname, node, launch=None, argv="",
                 remaps=None, conditions=None):
        Resource.__init__(self, config, rosname, conditions = conditions)
        self.node = node
        self.launch = launch
        self.argv = argv or ""
        self.remaps = remaps if not remaps is None else {}
        self.publishers = []
        self.subscribers = []
        self.servers = []
        self.clients = []
        self.reads = []
        self.writes = []
        self._location = launch.location if launch is not None else None
        self.location2 = loc_to_loc2(self._location)

    @property
    def type(self):
        return self.node.node_name

    @property
    def resource_type(self):
        return "node"

    @property
    def rt_outlinks(self):
        visited = set()
        queue = [self]
        nodes = []
        while queue:
            current = queue.pop(0)
            if current.id in visited:
                continue
            nodes.append(current)
            visited.add(current.id)
            for pub in self.publishers:
                for sub in pub.topic.subscribers:
                    queue.append(sub.node)
            for cli in self.clients:
                if cli.service.server:
                    queue.append(cli.service.server.node)
        return nodes

    def traceability(self):
        if self._location:
            return [self._location]
        return []

    def traceability2(self):
        return [self.location2]

    def remap(self, rosname):
        new = NodeInstance(self.configuration, rosname, self.node,
                           launch = self.launch, argv = list(self.argv),
                           remaps = dict(self.remaps),
                           conditions = list(self.conditions))
        new.publishers = list(self.publishers)
        new.subscribers = list(self.subscribers)
        new.servers = list(self.servers)
        new.clients = list(self.clients)
        new.reads = list(self.reads)
        new.writes = list(self.writes)
        new.location2 = self.location2
        return new

    def to_JSON_object(self):
        return {
            "uid": str(id(self)),
            "name": self.id,
            "type": self.node.node_name,
            "args": self.argv,
            "conditions": [c.to_JSON_object() for c in self.conditions],
            "publishers": [p.topic.rosname.full for p in self.publishers],
            "subscribers": [p.topic.rosname.full for p in self.subscribers],
            "servers": [p.service.rosname.full for p in self.servers],
            "clients": [p.service.rosname.full for p in self.clients],
            "reads": [p.parameter.rosname.full for p in self.reads],
            "writes": [p.parameter.rosname.full for p in self.writes],
            "traceability": [loc2_to_JSON(l) for l in self.traceability2()],
            "variables": self.variables
        }

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return ("NodeInstance " + self.configuration.name
                + ":" + self.rosname.full)


class Topic(Resource):
    def __init__(self, config, rosname, message_type = None, conditions = None):
        Resource.__init__(self, config, rosname, conditions = conditions)
        self.type = message_type
        self.publishers = []
        self.subscribers = []

    @property
    def is_disconnected(self):
        p = len(self.publishers)
        s = len(self.subscribers)
        return p + s > 0 and (p == 0 or s == 0)

    @property
    def resource_type(self):
        return "topic"

    @property
    def types(self):
        types = set(p.type for p in self.publishers)
        types.update(s.type for s in self.subscribers)
        return types

    def traceability(self):
        sl = []
        for p in self.publishers:
            if not p.source_location is None:
                sl.append(p.source_location)
        for p in self.subscribers:
            if not p.source_location is None:
                sl.append(p.source_location)
        return sl

    def traceability2(self):
        return [loc_to_loc2(l) for l in self.traceability()]

    def remap(self, rosname):
        new = Topic(self.configuration, rosname, message_type = self.type,
                    conditions = list(self.conditions))
        new.publishers = list(self.publishers)
        new.subscribers = list(self.subscribers)
        return new

    def to_JSON_object(self):
        return {
            "uid": str(id(self)),
            "name": self.id,
            "type": self.type,
            "conditions": [c.to_JSON_object() for c in self.conditions],
            "publishers": [p.node.rosname.full for p in self.publishers],
            "subscribers": [p.node.rosname.full for p in self.subscribers],
            "traceability": [loc2_to_JSON(l) for l in self.traceability2()],
            "variables": self.variables
        }

    def _get_conditions(self):
        conditional = True
        conditions = []
        for links in (self.publishers, self.subscribers):
            for link in links:
                if not link.node.conditions:
                    conditional = False
                    if not link.conditions:
                        return []
                conditions.extend(link.conditions)
                if conditional:
                    conditions.extend(link.node.conditions)
        return conditions


class Service(Resource):
    def __init__(self, config, rosname, message_type = None, conditions = None):
        Resource.__init__(self, config, rosname, conditions = conditions)
        self.type = message_type
        self.server = None
        self.clients = []

    @property
    def is_disconnected(self):
        s = 1 if not self.server is None else 0
        c = len(self.clients)
        return s + c > 0 and (s == 0 or c == 0)

    @property
    def servers(self):
        if self.server:
            return (self.server,)
        return ()

    @property
    def resource_type(self):
        return "service"

    @property
    def types(self):
        types = set(s.type for s in self.servers)
        types.update(c.type for c in self.clients)
        return types

    def traceability(self):
        sl = []
        if not self.server is None:
            if not self.server.source_location is None:
                sl.append(self.server.source_location)
        for p in self.clients:
            if not p.source_location is None:
                sl.append(p.source_location)
        return sl

    def traceability2(self):
        return [loc_to_loc2(l) for l in self.traceability()]

    def remap(self, rosname):
        new = Service(self.configuration, rosname, message_type = self.type,
                      conditions = list(self.conditions))
        new.server = self.servers
        new.clients = list(self.clients)
        return new

    def to_JSON_object(self):
        return {
            "uid": str(id(self)),
            "name": self.id,
            "type": self.type,
            "conditions": [c.to_JSON_object() for c in self.conditions],
            "servers": ([self.server.node.rosname.full]
                        if not self.server is None else []),
            "clients": [p.node.rosname.full for p in self.clients],
            "traceability": [loc2_to_JSON(l) for l in self.traceability2()],
            "variables": self.variables
        }

    def _get_conditions(self):
        conditional = True
        conditions = []
        for links in (self.servers, self.clients):
            for link in links:
                if not link.node.conditions:
                    conditional = False
                    if not link.conditions:
                        return []
                conditions.extend(link.conditions)
                if conditional:
                    conditions.extend(link.node.conditions)
        return conditions


class Parameter(Resource):
    def __init__(self, config, rosname, ptype, value,
                 node_scope = False, launch = None, conditions = None):
        Resource.__init__(self, config, rosname, conditions = conditions)
        self.type = ptype or Parameter.type_of(value)
        self.value = value
        self.node_scope = node_scope
        self.reads = []
        self.writes = []
        self.launch = launch
        self._location = launch.location if launch is not None else None
        self.location2 = loc_to_loc2(self._location)

    @staticmethod
    def type_of(value):
        if value is None:
            return None
        if value is True or value is False:
            return "bool"
        if isinstance(value, int):
            return "int"
        if isinstance(value, float):
            return "double"
        if isinstance(value, basestring):
            return "str"
        return "yaml"

    @property
    def resource_type(self):
        return "param"

    @property
    def types(self):
        types = set(r.type for r in self.reads)
        types.update(w.type for w in self.writes)
        return types

    def traceability(self):
        sl = []
        if self._location is not None:
            sl.append(self._location)
        for p in self.reads:
            if not p.source_location is None:
                sl.append(p.source_location)
        for p in self.writes:
            if not p.source_location is None:
                sl.append(p.source_location)
        return sl

    def traceability2(self):
        return [loc_to_loc2(l) for l in self.traceability()]

    def remap(self, rosname):
        new = Parameter(self.configuration, rosname, self.type,
                         self.value, node_scope = self.node_scope,
                         conditions = list(self.conditions))
        new.reads = list(self.reads)
        new.writes = list(self.writes)
        new.location2 = self.location2
        return new

    def to_JSON_object(self):
        return {
            "uid": str(id(self)),
            "name": self.id,
            "type": self.type,
            "value": self.value,
            "conditions": [c.to_JSON_object() for c in self.conditions],
            "reads": [p.node.rosname.full for p in self.reads],
            "writes": [p.node.rosname.full for p in self.writes],
            "traceability": [loc2_to_JSON(l) for l in self.traceability2()],
            "variables": self.variables
        }


class ResourceCollection(object):
    def __init__(self, iterable):
        self.all = []
        self.enabled = []
        self.unresolved = []
        self.conditional = []
        self.counter = Counter()
        if not iterable is None:
            for resource in iterable:
                self.add(resource)

    def __len__(self):
        return len(self.all)

    def __getitem__(self, item):
        return self.all[item]

    def __contains__(self, key):
        return key in self.counter

    def __iter__(self):
        return self.all.__iter__()

    def get(self, name, conditional = True):
        for resource in reversed(self.all):
            if resource.id == name:
                if conditional or not resource.conditions:
                    return resource
        return None

    def get_all(self, name, conditional = True):
        resources = []
        for resource in self.all:
            if resource.id == name:
                if conditional or not resource.conditions:
                    resources.append(resource)
        return resources

    def get_collisions(self):
        return len(self.all) - len(self.counter)

    def add(self, resource):
        self.all.append(resource)
        if resource.conditions:
            self.conditional.append(resource)
        else:
            self.enabled.append(resource)
        if "?" in resource.id:
            self.unresolved.append(resource)
        previous = self.counter[resource.id]
        self.counter[resource.id] += 1
        return previous

    def remove(self, name):
        for col in (self.all, self.conditional, self.enabled, self.unresolved):
            for i in range(len(col) - 1, -1, -1):
                if col[i].id == name:
                    del col[i]
        previous = 0
        if name in self.counter:
            previous = self.counter[name]
            self.counter[name] = 0
        return previous


LaunchCommand = namedtuple("LaunchCommand", ("command", "args"))


class Configuration(MetamodelObject):
    """A configuration is more or less equivalent to an application.
        It is the result of a set of launch files,
        plus environment, parameters, etc.
    """

    def __init__(self, name, env = None, nodes = None,
                 topics = None, services = None, parameters = None):
        self.id = "configuration:" + name
        self.name = name
        self.roslaunch = []
        self.environment = env if not env is None else {}
        self.nodes = ResourceCollection(nodes)
        self.topics = ResourceCollection(topics)
        self.services = ResourceCollection(services)
        self.parameters = ResourceCollection(parameters)
        self.dependencies = DependencySet()
        self.launch_commands = [] # [LaunchCommand]
        self.hpl_properties = [] # [string | HplAstObject]
        self.hpl_assumptions = [] # [string | HplAstObject]
        self.user_attributes = {}

    @property
    def location(self):
        return RuntimeLocation(self)

    def get_collisions(self):
        counter = Counter()
        counter += self.nodes.counter
        counter += self.topics.counter
        counter += self.services.counter
        counter += self.parameters.counter
        return sum(counter.values()) - len(counter)

    def get_remaps(self):
        unique = set()
        for node in self.nodes:
            unique.update(node.remaps.viewitems())
        return len(unique)

    def get_unresolved(self):
        return (len(self.nodes.unresolved) + len(self.topics.unresolved)
                + len(self.services.unresolved)
                + len(self.parameters.unresolved))

    def get_conditional(self):
        # FIXME len(self.topics.conditional) does not always work
        n = 0
        for c in (self.nodes, self.topics, self.services, self.parameters):
            for r in c:
                if r.conditions:
                    n += 1
        return n

    def add_command(self, cmd, args):
        self.launch_commands.append(LaunchCommand(cmd, args))

    def to_JSON_object(self):
        publishers = []
        subscribers = []
        servers = []
        clients = []
        reads = []
        writes = []
        for node in self.nodes:
            publishers.extend(p.to_JSON_object() for p in node.publishers)
            subscribers.extend(p.to_JSON_object() for p in node.subscribers)
            servers.extend(p.to_JSON_object() for p in node.servers)
            clients.extend(p.to_JSON_object() for p in node.clients)
            reads.extend(p.to_JSON_object() for p in node.reads)
            writes.extend(p.to_JSON_object() for p in node.writes)
        return {
            "id": self.name,
            "launch": [f.to_JSON_object() for f in self.roslaunch],
            "collisions": self.get_collisions(),
            "remaps": self.get_remaps(),
            "dependencies": list(self.dependencies.packages),
            "environment": list(self.dependencies.environment),
            "nodes": [n.to_JSON_object() for n in self.nodes],
            "topics": [t.to_JSON_object() for t in self.topics],
            "services": [s.to_JSON_object() for s in self.services],
            "parameters": [p.to_JSON_object() for p in self.parameters],
            "links": {
                "publishers": publishers,
                "subscribers": subscribers,
                "servers": servers,
                "clients": clients,
                "reads": reads,
                "writes": writes
            },
            "hpl": {
                "properties": map(str, self.hpl_properties),
                "assumptions": map(str, self.hpl_assumptions)
            }
        }

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Configuration " + self.name


###############################################################################
# ROS Runtime Analysis Properties
###############################################################################

class RosPrimitive(MetamodelObject):
    def __init__(self, node, rosname, conditions = None, location = None):
        self.node = node
        self.rosname = rosname # before remappings
        self.conditions = conditions if not conditions is None else []
        self.source_location = location
        self.location2 = loc_to_loc2(location)

    @property
    def location(self):
        return self.node.location

    @property
    def configuration(self):
        return self.node.configuration

    def to_JSON_object(self):
        return {
            "node": self.node.rosname.full,
            "node_uid": str(id(self.node)),
            "name": self.rosname.full,
            "location": loc2_to_JSON(self.location2),
            "conditions": [c.to_JSON_object() for c in self.conditions]
        }


class TopicPrimitive(RosPrimitive):
    def __init__(self, node, topic, message_type, rosname, queue_size,
                 conditions = None, location = None):
        RosPrimitive.__init__(self, node, rosname, conditions = conditions,
                              location = location)
        self.topic = topic
        self.type = message_type
        self.queue_size = queue_size

    @property
    def topic_name(self):
        return self.topic.rosname.full

    def to_JSON_object(self):
        data = RosPrimitive.to_JSON_object(self)
        data["topic"] = self.topic_name
        data["topic_uid"] = str(id(self.topic))
        data["type"] = self.type
        data["queue"] = self.queue_size
        return data

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Link of node '{}' to topic '{}' of type '{}'".format(
            self.node.id, self.topic.id, self.type)

class PublishLink(TopicPrimitive):
    @classmethod
    def link(cls, node, topic, message_type, rosname, queue_size,
             latched=False, conditions=None, location=None):
        link = cls(node, topic, message_type, rosname, queue_size,
                   conditions = conditions, location = location)
        link.node.publishers.append(link)
        link.topic.publishers.append(link)
        link.latched = latched
        return link

    @classmethod
    def link_from_call(cls, node, topic, rosname, call):
        if not isinstance(call, AdvertiseCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, topic, call.type, rosname, call.queue_size,
                   conditions=list(call.conditions), location=call.location)
        link.node.publishers.append(link)
        link.topic.publishers.append(link)
        link.latched = call.latched
        link.location2 = call.location2
        return link

    def to_JSON_object(self):
        data = TopicPrimitive.to_JSON_object(self)
        data["latched"] = self.latched
        return data

    def __str__(self):
        return "Publication of node '{}' to topic '{}' of type '{}'".format(
            self.node.id, self.topic.id, self.type)

class SubscribeLink(TopicPrimitive):
    @classmethod
    def link(cls, node, topic, message_type, rosname, queue_size,
             conditions = None, location = None):
        link = cls(node, topic, message_type, rosname, queue_size,
                   conditions = conditions, location = location)
        link.node.subscribers.append(link)
        link.topic.subscribers.append(link)
        return link

    @classmethod
    def link_from_call(cls, node, topic, rosname, call):
        if not isinstance(call, SubscribeCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, topic, call.type, rosname, call.queue_size,
                   conditions=list(call.conditions), location=call.location)
        link.node.subscribers.append(link)
        link.topic.subscribers.append(link)
        link.location2 = call.location2
        return link

    def __str__(self):
        return "Subscription of node '{}' to topic '{}' of type '{}'".format(
            self.node.id, self.topic.id, self.type)


class ServicePrimitive(RosPrimitive):
    def __init__(self, node, service, message_type, rosname,
                 conditions = None, location = None):
        RosPrimitive.__init__(self, node, rosname, conditions = conditions,
                              location = location)
        self.service = service
        self.type = message_type

    @property
    def topic_name(self):
        return self.service.rosname.full

    def to_JSON_object(self):
        data = RosPrimitive.to_JSON_object(self)
        data["service"] = self.topic_name
        data["service_uid"] = str(id(self.service))
        data["type"] = self.type
        return data

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "SrvCli({}, {}, {})".format(self.node.id, self.service.id,
                                           self.type)

class ServiceLink(ServicePrimitive):
    @classmethod
    def link(cls, node, service, message_type, rosname, conditions = None,
             location = None):
        link = cls(node, service, message_type, rosname,
                   conditions = conditions, location = location)
        link.node.servers.append(link)
        link.service.server = link
        return link

    @classmethod
    def link_from_call(cls, node, service, rosname, call):
        if not isinstance(call, AdvertiseServiceCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, service, call.type, rosname,
                   conditions=list(call.conditions), location=call.location)
        link.node.servers.append(link)
        link.service.server = link
        link.location2 = call.location2
        return link

    def __str__(self):
        return "Service({}, {}, {})".format(self.node.id, self.service.id,
                                            self.type)

class ClientLink(ServicePrimitive):
    @classmethod
    def link(cls, node, service, message_type, rosname, conditions = None,
             location = None):
        link = cls(node, service, message_type, rosname,
                   conditions = conditions, location = location)
        link.node.clients.append(link)
        link.service.clients.append(link)
        return link

    @classmethod
    def link_from_call(cls, node, service, rosname, call):
        if not isinstance(call, ServiceClientCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, service, call.type, rosname,
                   conditions=list(call.conditions), location=call.location)
        link.node.clients.append(link)
        link.service.clients.append(link)
        link.location2 = call.location2
        return link

    def __str__(self):
        return "Client({}, {}, {})".format(self.node.id, self.service.id,
                                           self.type)


class ParameterPrimitive(RosPrimitive):
    def __init__(self, node, param, param_type, rosname, value=None,
                 conditions=None, location=None):
        RosPrimitive.__init__(self, node, rosname, conditions = conditions,
                              location = location)
        self.parameter = param
        self.type = param_type
        self.value = value

    @property
    def param_name(self):
        return self.parameter.rosname.full

    def to_JSON_object(self):
        data = RosPrimitive.to_JSON_object(self)
        data["param"] = self.param_name
        data["param_uid"] = str(id(self.parameter))
        data["type"] = self.type
        data["value"] = self.value
        return data

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Param({}, {}, {})".format(self.node.id, self.parameter.id,
                                          self.type)

class ReadLink(ParameterPrimitive):
    @classmethod
    def link(cls, node, param, param_type, rosname, value=None,
             conditions=None, location=None):
        link = cls(node, param, param_type, rosname, value=value,
                   conditions=conditions, location=location)
        link.node.reads.append(link)
        link.parameter.reads.append(link)
        return link

    @classmethod
    def link_from_call(cls, node, param, rosname, call):
        if not isinstance(call, GetParamCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, param, call.type, rosname, value=call.default_value,
                   conditions=list(call.conditions), location=call.location)
        link.node.reads.append(link)
        link.parameter.reads.append(link)
        link.location2 = call.location2
        return link

    def __str__(self):
        return "Read({}, {}, {})".format(self.node.id, self.parameter.id,
                                         self.type)

class WriteLink(ParameterPrimitive):
    @classmethod
    def link(cls, node, param, param_type, rosname, value=None,
             conditions=None, location=None):
        link = cls(node, param, param_type, rosname, conditions = conditions,
                   location = location)
        link.node.writes.append(link)
        link.parameter.writes.append(link)
        return link

    @classmethod
    def link_from_call(cls, node, param, rosname, call):
        if not isinstance(call, SetParamCall):
            raise ValueError("wrong call type: " + type(call).__name__)
        link = cls(node, param, call.type, rosname, value=call.value,
                   conditions=list(call.conditions), location=call.location)
        link.node.writes.append(link)
        link.parameter.writes.append(link)
        link.location2 = call.location2
        return link

    def __str__(self):
        return "Write({}, {}, {})".format(self.node.id, self.parameter.id,
                                          self.type)


###############################################################################
# Helper Functions
###############################################################################

def _cpp_ignore_line(line):
    return "// haros:ignore-line" in line

def _cpp_ignore_next_line(line):
    return "// haros:ignore-next-line" in line

def _py_ignore_line(line):
    return "# haros:ignore-line" in line

def _py_ignore_next_line(line):
    return "# haros:ignore-next-line" in line

def _no_parser(line):
    return False

def _bool_to_conditions(v):
    v = bool(v)
    if v is True:
        return []
    else:
        return [False]

def fpid(v):
    return v


def loc_to_loc2(loc):
    if loc is None or loc.package is None:
        return Location2(None, None, None, None)
    pkg = loc.package.name
    if loc.file is None:
        return Location2(pkg, None, None, None)
    fname = loc.file.full_name
    if loc.line is None:
        return Location2(pkg, fname, None, None)
    return Location2(pkg, fname, loc.line, loc.column)

def loc2_to_JSON(loc2):
    return {
        "package": loc2.package,
        "file": loc2.file,
        "line": loc2.line,
        "column": loc2.column,
        "function": None,
        "class": None
    }

def JSON_to_loc2(data):
    if data is None:
        return Location2(None, None, None, None)
    pkg = data.get("package")
    if pkg is None:
        return Location2(None, None, None, None)
    fname = data.get("file")
    if fname is None:
        return Location2(pkg, None, None, None)
    line = data.get("line")
    if line is None:
        return Location2(pkg, fname, None, None)
    return Location2(pkg, fname, line, data.get("column"))


###############################################################################
# Test Functions
###############################################################################

def test_rosname():
    n1 = RosName("a")
    n2 = RosName("a", "ns")
    assert n1 != n2
    assert n1 == "/a"
    assert "/a" == n1
    assert n2 == "ns/a"
    assert "ns/a" == n2
    n1 = RosName("a", "ns")
    assert n1 == n2
    n1 = RosName("a")
    n2 = RosName("a")
    assert n1 == n2
    n1 = RosName("~a", "ns", "priv")
    n2 = RosName("a", "ns")
    assert n1 != n2
    assert n1 == "priv/a"
    n = Node("base", Package("pkg"), rosname = RosName("base"))
    assert n.rosname == "/base"



if __name__ == "__main__":
    test_rosname()



# -*- coding: utf-8 -*-

#Copyright (c) 2017 André Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

###############################################################################
# Notes to Self
###############################################################################

# When parsing launch files, unresolved values could be tolerated, because
# the parse tree needs to be generic. In this case, when building a
# configuration, we must have all information available, including environment.
# As such, when traversing launch trees, if we are unable to resolve a value,
# it should be considered a configuration error and stored as such.
# The only exception is conditionals. If we are unable to resolve a
# conditional, the resulting entity should be marked as a conditional entity.


###############################################################################
# Imports
###############################################################################

from builtins import range
from collections import namedtuple
from itertools import chain
import logging
import os
import re
import yaml

rosparam = None # lazy import

from .extractor import HardcodedNodeParser, PackageExtractor
from .launch_parser import (
    LaunchParser, LaunchParserError, SubstitutionError, SubstitutionParser
)
from .metamodel import (
    Node, Configuration, RosName, NodeInstance, Parameter, Topic, Service,
    SourceCondition, PublishLink, SubscribeLink, ServiceLink, ClientLink,
    ReadLink, WriteLink,
    AdvertiseCall, SubscribeCall, AdvertiseServiceCall, ServiceClientCall,
    GetParamCall, SetParamCall, Location2, JSON_to_loc2, _bool_to_conditions
)


###############################################################################
# Utility
###############################################################################

class LoggingObject(object):
    log = logging.getLogger(__name__)


_EMPTY_DICT = {}
_EMPTY_LIST = ()


###############################################################################
# Launch File Analysis
###############################################################################

class LaunchScope(LoggingObject):
    TempParam = namedtuple("TempParam", ["name", "type", "value", "ifs"])

    def __init__(self, parent, config, launch_file, ns="/", node=None,
                 remaps=None, params=None, args=None, conditions=None):
        self.parent = parent
        self.children = []
        self.configuration = config
        self.launch_file = launch_file
        self.namespace = ns
        self.node = node
        self.remaps = remaps if not remaps is None else {}
        self.parameters = params if not params is None else []
        self.arguments = args if not args is None else {}
        self.conditions = conditions if not conditions is None else []
        self._params = list(parent._params) if parent else []

    @property
    def private_ns(self):
        return self.node.rosname.full if self.node else self.namespace

    def child(self, ns, condition, launch = None, args = None):
        launch = launch or self.launch_file
        new = LaunchScope(self, self.configuration, launch,
                          ns = self._namespace(ns),
                          remaps = dict(self.remaps),
                          params = self.parameters,
                          args = args if not args is None else self.arguments,
                          conditions = list(self.conditions))
        if not condition is True:
            new.conditions.append(condition)
        self.children.append(new)
        return new

    def remap(self, source, target):
        pns = self.private_ns
        source = RosName.resolve(source, self.namespace, private_ns = pns)
        target = RosName.resolve(target, self.namespace, private_ns = pns)
        self.remaps[source] = target

    def make_node(self, node, name, ns, args, condition, line=None, col=None):
        ns = self._namespace(ns)
        name = name or node.rosname.own
        rosname = RosName(name, ns, self.private_ns)
        self.log.debug("Creating NodeInstance %s for Node %s.",
                       rosname.full, node.name)
        instance = NodeInstance(self.configuration, rosname, node,
            launch=self.launch_file, argv = args, remaps=dict(self.remaps),
            conditions=list(self.conditions))
        if instance._location is not None:
            instance._location.line = line
            instance._location.column = col
        node.instances.append(instance)
        if not condition is True:
            instance.conditions.append(condition)
        previous = self.configuration.nodes.add(instance)
        new_scope = LaunchScope(self, self.configuration, self.launch_file,
            ns=ns, node=instance, remaps=instance.remaps,
            params=self.parameters, args=self.arguments,
            conditions=instance.conditions)
        self.children.append(new_scope)
        pns = new_scope.private_ns
        for param in self._params:
            rosname = RosName(param.rosname.given, pns, pns)
            conditions = param.conditions + instance.conditions
            self.log.debug("Creating new forward Parameter %s.", rosname.full)
            new_param = Parameter(self.configuration, rosname, param.type,
                param.value, node_scope=param.node_scope,
                launch=param.launch_file, conditions=conditions)
            new_param._location = param._location
            self.parameters.append(new_param)
        return new_scope

    def make_params(self, name, ptype, value, condition, line=None, col=None):
        if not value is None:
            value = self._convert_value(str(value), ptype)
            ptype = Parameter.type_of(value)
        conditions = list(self.conditions)
        if not condition is True:
            conditions.append(condition)
        if ptype == "yaml" or isinstance(value, dict):
            self._yaml_param(name, value, conditions, line=line, col=col)
        else:
            rosname = RosName(name, self.private_ns, self.private_ns)
            param = Parameter(self.configuration, rosname, ptype, value,
                              node_scope = not self.node is None,
                              launch = self.launch_file,
                              conditions = conditions)
            if param._location is not None:
                param._location.line = line
                param._location.column = col
            if not self.node and rosname.is_private:
                self._add_param(param, self._params)
            else:
                self._add_param(param, self.parameters)

    def make_rosparam(self, name, ns, value, condition, line=None, col=None):
    # ---- lazy rosparam import as per the oringinal roslaunch code
        global rosparam
        if rosparam is None:
            import rosparam
        try:
            value = yaml.safe_load(value)
        except yaml.MarkedYAMLError as e:
            raise ConfigurationError(str(e))
    # ----- try to use given name, namespace or both
        ns = self._ns_join(ns or self.private_ns, self.private_ns)
        if name:
            name = self._ns_join(name, ns)
        else:
            if not isinstance(value, dict):
                raise ConfigurationError("'param' attribute must be set"
                                         " for non-dictionary values")
    # ----- this will unfold, so we can use namespace in place of a name
            name = ns
        conditions = list(self.conditions)
        if not condition is True:
            conditions.append(condition)
        self._yaml_param(name, value, conditions, private=False,
                         line=line, col=col)

    def remove_param(self, name, ns, condition):
        # TODO check whether "~p" = "/rosparam/p" is intended or a bug
        ns = self._ns_join(ns or self.private_ns, self.private_ns)
        name = RosName.resolve(name, ns, "/rosparam")
        param = self.configuration.parameters.get(name)
        if not param:
            raise ConfigurationError("missing parameter: " + name)
        if not condition is True or self.conditions:
            return
        else:
            self.resources.deleted_params.append(name)

    def _namespace(self, ns, private = False):
        pns = self.private_ns
        if not ns:
            return self.namespace if not private else pns
        if private:
            return RosName.resolve(ns, pns, private_ns = pns)
        return RosName.resolve(ns, self.namespace)

    # as seen in roslaunch code, sans a few details
    def _convert_value(self, value, ptype):
        if ptype is None:
            # attempt numeric conversion
            try:
                if "." in value:
                    return float(value)
                else:
                    return int(value)
            except ValueError as e:
                pass
            # bool
            lval = value.lower()
            if lval == "true" or lval == "false":
                return self._convert_value(value, "bool")
            # string
            return value
        elif ptype == "str" or ptype == "string":
            return value
        elif ptype == "int":
            return int(value)
        elif ptype == "double":
            return float(value)
        elif ptype == "bool":
            value = value.lower().strip()
            if value == "true" or value == "1":
                return True
            elif value == "false" or value == "0":
                return False
            raise ValueError("{} is not a '{}' type".format(value, ptype))
        elif ptype == "yaml":
            try:
                return yaml.safe_load(value)
            except yaml.parser.ParserError as e:
                raise ValueError(e)
        else:
            raise ValueError("Unknown type '{}'".format(ptype))

    def _yaml_param(self, name, value, conditions, private=True,
                    line=None, col=None):
        private = private and name.startswith("~")
        pns = self.private_ns
        node_scope = not self.node is None
        items = self._unfold(name, value)
        for name, value, independent in items:
            independent = name.startswith("/") or name.startswith("~")
            if independent and name.startswith("~"):
                rosname = RosName(name, "/roslaunch", "/roslaunch")
            else:
                rosname = RosName(name, pns, pns)
            param = Parameter(self.configuration, rosname, None, value,
                              node_scope = node_scope,
                              launch = self.launch_file, conditions = conditions)
            if param._location is not None:
                param._location.line = line
                param._location.column = col
            if independent or not private:
                self._add_param(param, self.parameters)
            else:
                self._add_param(param, self._params)

    def _unfold(self, name, value):
        result = []
        stack = [("", name, value)]
        while stack:
            ns, key, value = stack.pop()
            name = self._ns_join(key, ns)
            if not isinstance(value, dict):
                result.append((name, value, name == key)) #FIXME sometimes not independent: ~ns/p + a != a
            else:
                for key, other in value.iteritems():
                    stack.append((name, key, other))
        return result

    def _ns_join(self, name, ns):
        """Dumb version of name resolution to mimic ROS behaviour."""
        if name.startswith("~") or name.startswith("/"):
            return name
        if ns == "~":
            return "~" + name
        if not ns:
            return name
        if ns[-1] == "/":
            return ns + name
        return ns + "/" + name

    def _add_param(self, param, collection):
        if param.rosname.is_unresolved:
            collection.append(param)
        else:
            rosname = param.rosname.full
            for i in range(len(collection)):
                other = collection[i]
                if rosname == other.rosname.full:
                    if param.disabled:
                        if other.disabled:
                            collection[i] = param
                    else:
                        collection[i] = param
                    return
            collection.append(param)


###############################################################################
# Configuration Builder
###############################################################################

class ConfigurationError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class ConfigurationBuilder(LoggingObject):
    def __init__(self, name, environment, source_finder,
                 hints=None, no_hardcoded=False):
        self.configuration = Configuration(name, env = environment)
        self.sources = source_finder
        self.errors = []
        self.no_hardcoded = no_hardcoded
        self._future = []
        self._pkg_finder = PackageExtractor() # FIXME should this be given?
        self._pkg_finder.packages.extend(self.sources.packages.values())
        hints = hints or {}
        self._fix_hints = hints.get("fix", _EMPTY_DICT)
        #self._add_nodes = hints.get("nodes", _EMPTY_DICT)
        #self._add_params = hints.get("params", _EMPTY_DICT)

        self.hints = ConfigHints2(hints, self._get_node)
        self._invalid = False

    def add_rosrun(self, node):
        if self._invalid:
            raise ConfigurationError("invalid state")
        config = self.configuration
        self.log.debug("Adding rosrun command to configuration. Node: %s",
                       node.node_name)
        if node.is_nodelet:
            raise ValueError("Cannot add a nodelet via 'rosrun' configuration.")
            # ^ because in this case we do not have the args
        name = node.name if not node.rosname else node.rosname.own
        config.add_command("rosrun", [node.package.name, node.name, name])
        scope = LaunchScope(None, config, None)
        scope = scope.make_node(node, name, "/", (), True)
        self._future.append(FutureNodeLinks(scope.node))

    def add_launch(self, launch_file):
        if self._invalid:
            raise ConfigurationError("invalid state")
        assert launch_file.language == "launch"
        config = self.configuration
        config.roslaunch.append(launch_file)
        config.add_command("roslaunch", [launch_file.path])
        if not launch_file.tree:
            self.errors.append("missing parse tree: " + launch_file.id)
            return False
        sub = SubstitutionParser(env=config.environment,
            pkgs=self.sources.packages, dirname=launch_file.dir_path,
            pkg_depends=config.dependencies.packages,
            env_depends=config.dependencies.environment)
        scope = LaunchScope(None, self.configuration, launch_file,
                            args = sub.arguments)
        self._analyse_tree(launch_file.tree, scope, sub)
    # ----- parameters can only be added in the end, because of rosparam
        for param in scope.parameters:
            self.configuration.parameters.add(param)

    def build(self):
        if not self._invalid:
            # finishing touches
            # BY FIRE BE PURGED
            # fix params: harmless
            self.hints.apply_param_fixes(self.configuration)
            # add params: harmless
            self.hints.create_new_params(self.configuration)
            # fix nodes: can use new params
            self.hints.apply_node_fixes(self.configuration, self._future)
            # make links after fixing stuff
            for future_node_links in self._future:
                future_node_links.make(hints=self._fix_hints)
            # reset futures and add more nodes and links
            self._future = []
            self.hints.create_new_nodes(self.configuration, self._future)
            # make links again for the new nodes
            for future_node_links in self._future:
                future_node_links.make(hints=self._fix_hints)
            self._future = []
            # fix conditions from links
            self._update_topic_conditions()
            self._update_service_conditions()
            self._update_param_conditions()
            self._invalid = True
        return self.configuration

    def _analyse_tree(self, tree, scope, sub):
        for tag in tree.children:
            if tag.tag == "error":
                self.errors.append(tag.text)
                continue
            try:
                condition = self._condition(tag.condition, sub,
                                            tag.line, tag.column)
                if condition is False:
                    continue
                handler = getattr(self, "_" + tag.tag + "_tag")
                handler(tag, condition, scope, sub)
            except (ConfigurationError, SubstitutionError) as e:
                self.errors.append(e.value)

    def _node_tag(self, tag, condition, scope, sub):
        pkg = sub.resolve(tag.package, strict = True)
        exe = sub.resolve(tag.type, strict = True)
        args = sub.resolve(tag.argv, strict = True)
        if not pkg or not exe:
            raise ConfigurationError("node tag is missing pkg or type")
        node = self._get_node(pkg, exe, args)
        name = sub.resolve(tag.name, strict = True)
        ns = sub.resolve(tag.namespace, strict = True)
        new_scope = scope.make_node(node, name, ns, args, condition,
                                    line=tag.line, col=tag.column)
        self._future.append(FutureNodeLinks(new_scope.node))
        self._analyse_tree(tag, new_scope, sub)

    def _include_tag(self, tag, condition, scope, sub):
        filepath = sub.resolve(tag.file, strict = True)
        ns = sub.resolve(tag.namespace, strict = True)
        pass_all_args = sub.resolve(tag.pass_all_args, strict = True)
        self.log.debug("<include> " + str(filepath))
        launch_file = self.sources.get_file(filepath)
        if launch_file is None:
            launch_file = self._lookup_launch(filepath)
        if not launch_file.tree:
            self._parse_launch_on_the_fly(launch_file)
            if not launch_file.tree:
                self.log.debug("unable to parse launch: " + filepath)
                raise ConfigurationError("cannot parse: " + launch_file.id)
        else:
            self.log.debug("launch file '%s' is pre-parsed", launch_file.name)
        args = dict(scope.arguments) if pass_all_args else {}
        new_scope = scope.child(ns, condition, launch = launch_file,
                                args = args)
        # define child args in the new scope
        self._analyse_tree(tag, new_scope, sub)
        new_sub = SubstitutionParser(args = args, env = sub.environment,
                                     pkgs = sub.packages,
                                     dirname = launch_file.dir_path,
                                     pkg_depends = sub.pkg_depends,
                                     env_depends = sub.env_depends)
        self.log.debug("analyse_tree('%s')", launch_file.name)
        self._analyse_tree(launch_file.tree, new_scope, new_sub)

    def _remap_tag(self, tag, condition, scope, sub):
        assert not tag.children
        if not condition is True:
            self.errors.append("cannot resolve conditional remap")
        else:
            origin = sub.resolve(tag.origin, strict = True)
            target = sub.resolve(tag.target, strict = True)
            scope.remap(origin, target)

    def _param_tag(self, tag, condition, scope, sub):
        assert not tag.children
        name = sub.resolve(tag.name, strict=True)
        ptype = sub.resolve(tag.type, strict=True)
        if not tag.value is None:
            value = sub.resolve(tag.value, strict=True)
        elif not tag.textfile is None:
            try:
                with open(tag.textfile) as f:
                    value = f.read()
            except IOError as e:
                raise ConfigurationError("cannot read file: " + tag.textfile)
        elif not tag.binfile is None:
            value = None
        elif not tag.command is None:
            value = None
        try:
            scope.make_params(name, ptype, value, condition,
                              line=tag.line, col=tag.column)
        except ValueError as e:
            raise ConfigurationError(str(e))

    def _rosparam_tag(self, tag, condition, scope, sub):
        assert not tag.children
        command = sub.resolve(tag.command, strict = True)
        ns = sub.resolve(tag.namespace, strict = True)
        filepath = sub.resolve(tag.file, strict = True)
        name = sub.resolve(tag.name, strict = True)
        if command == "load":
            if filepath:
                try:
                    with open(filepath) as f:
                        value = f.read()
                except IOError as e:
                    raise ConfigurationError("cannot read file: " + filepath)
            else:
                value = tag.text
                if sub.resolve(tag.substitute, strict = True):
                    value = sub.sub(value)
                    value = sub.resolve(value, strict = True)
            scope.make_rosparam(name, ns, value, condition,
                                line=tag.line, col=tag.column)
        elif command == "delete":
            scope.remove_param(name, ns, condition)

    def _group_tag(self, tag, condition, scope, sub):
        ns = sub.resolve(tag.namespace, strict = True)
        new_scope = scope.child(ns, condition)
        self._analyse_tree(tag, new_scope, sub)

    def _arg_tag(self, tag, condition, scope, sub):
        assert not tag.children
        if scope.conditions or not condition is True:
            raise ConfigurationError("cannot resolve conditional arg")
        if not tag.value is None:
            value = sub.resolve(tag.value, strict = True)
            scope.arguments[tag.name] = value
        else:
            value = sub.resolve(tag.default, strict = True)
            if scope.arguments.get(tag.name) is None:
                scope.arguments[tag.name] = value

    def _env_tag(self, tag, condition, scope, sub):
        assert not tag.children

    def _machine_tag(self, tag, condition, scope, sub):
        assert not tag.children

    def _test_tag(self, tag, condition, scope, sub):
        pass

    def _condition(self, condition, sub, line, col):
        assert isinstance(condition, tuple)
        value = sub.resolve(condition[1], conversion=bool)
        if value is None:
            stmt = "if" if condition[0] else "unless"
            loc = self.configuration.roslaunch[-1].location
            loc.line = line
            loc.column = col
            # not sure if tag is part of the configuration
            return SourceCondition(condition[1], # UnresolvedValue
                location=loc, statement=stmt)
        if value is condition[0]:
            # tag is part of the configuration
            return True
        # tag is not part of the configuration
        return False

    def _get_node(self, pkg, exe, args):
        if exe == "nodelet":
            if not args:
                raise ConfigurationError("nodelet without args")
            args = args.split()
            if args[0] == "load" or args[0] == "standalone":
                if len(args) < 3:
                    raise ConfigurationError("nodelet load: too few arguments")
                pkg, exe = args[1].split("/")

        node = self.sources.nodes.get("node:" + pkg + "/" + exe)
        package = self.sources.packages.get("package:" + pkg)
        if not package:
            assert not node
            if self.no_hardcoded:
                self.log.debug(("skipping hard-coded node '%s/%s' "
                                "due to user option"), pkg, exe)
            else:
                self.log.debug("look up hard-coded node '%s/%s'", pkg, exe)
                node = HardcodedNodeParser.get(pkg, exe)
            if not node:
                package = self._find_package(pkg)
        if not node:
            node = Node(exe, package, rosname = RosName("?"), nodelet = exe)
        return node

    def _find_package(self, name):
        # FIXME this is a hammer
        for pkg in self._pkg_finder.packages:
            if pkg.name == name:
                return pkg
        pkg = self._pkg_finder.find_package(name)
        if pkg is None:
            raise ConfigurationError("cannot find package: " + name)
        return pkg

    def _find_package_by_path(self, filepath):
        prev = filepath
        curr = os.path.dirname(filepath)
        while prev != curr:
            pkg = self._pkg_finder.find_package_at(curr, populate=True)
            if pkg is not None:
                return pkg
            prev = curr
            curr = os.path.dirname(curr)
        raise ConfigurationError("cannot find package for file: " + filepath)

    def _lookup_launch(self, filepath):
        self.log.debug("dynamic lookup of launch file: " + filepath)
        for pkg in chain(self._pkg_finder.packages, self._pkg_finder._extra):
            if not pkg.path:
                continue
            if (filepath.startswith(pkg.path)
                    and filepath[len(pkg.path)] == os.path.sep):
                self.log.debug("found package '%s' for launch file", pkg.name)
                break
        else:
            pkg = self._find_package_by_path(filepath)
            # FIXME we could just open the file at the given path, but then
            # we would not have the Package, File, Location objects.
            # The metamodel needs to be changed.
            #self.log.debug("failed to find package for launch file")
            #raise ConfigurationError("cannot find launch file: " + filepath)
        for sf in pkg.source_files:
            if sf.path == filepath:
                self.log.debug("found SourceFile '%s' for launch file", sf.name)
                break
        else:
            self.log.debug("failed to find SourceFile object for launch file")
            raise ConfigurationError("cannot find launch file: " + filepath)
        return sf

    def _parse_launch_on_the_fly(self, launch_file):
        assert not launch_file.tree
        assert launch_file.language == "launch"
        launch_parser = LaunchParser(pkgs=self._pkg_finder)
        self.log.debug("Parsing launch file: " + launch_file.path)
        try:
            launch_file.tree = launch_parser.parse(launch_file.path)
        except LaunchParserError as e:
            self.log.warning("Parsing error in %s:\n%s",
                             launch_file.path, str(e))

    def _update_topic_conditions(self):
        for topic in self.configuration.topics:
            assert not topic.conditions, "{!r} {!r}".format(
                topic.id, topic.conditions)
            for link in topic.publishers:
                if not link.conditions:
                    break
                topic.conditions.extend(link.conditions)
            else:
                for link in topic.subscribers:
                    if not link.conditions:
                        topic.conditions = []
                        break
                    topic.conditions.extend(link.conditions)

    def _update_service_conditions(self):
        for service in self.configuration.services:
            assert not service.conditions
            link = service.server
            if link:
                if not link.conditions:
                    continue
                else:
                    service.conditions.extend(link.conditions)
            for link in service.clients:
                if not link.conditions:
                    service.conditions = []
                    break
                service.conditions.extend(link.conditions)

    def _update_param_conditions(self):
        for param in self.configuration.parameters:
            if param.launch is not None:
                if not param.conditions:
                    continue
            for link in param.reads:
                if not link.conditions:
                    param.conditions = []
                    break
                param.conditions.extend(link.conditions)
            else:
                for link in param.writes:
                    if not link.conditions:
                        param.conditions = []
                        break
                    param.conditions.extend(link.conditions)


class FutureNodeLinks(LoggingObject):
    __slots__ = ("node", "ns", "pns")

    def __init__(self, node):
        self.node = node
        self.ns = node.namespace
        self.pns = node.rosname.full

        self.advertise = []
        self.subscribe = []
        self.service = []
        self.client = []
        self.read_param = []
        self.write_param = []

    _LINKS = (
        ("advertise", PublishLink, Topic, "topics"),
        ("subscribe", SubscribeLink, Topic, "topics"),
        ("service", ServiceLink, Service, "services"),
        ("client", ClientLink, Service, "services"),
        ("read_param", ReadLink, Parameter, "parameters"),
        ("write_param", WriteLink, Parameter, "parameters")
    )

    def make(self, hints=None):
        hints = hints or _EMPTY_DICT
        config = self.node.configuration
        # TODO *add* (not fix) publishers etc. from hints
        for node_attr, link_cls, rcls, col in self._LINKS:
            calls = getattr(self.node.node, node_attr)
            if calls:
                self.log.debug("Iterating %s calls for node %s.",
                    calls[0].KEY, self.node.id)
            collection = getattr(config, col)
            new_calls = [call.clone() for call in calls]
            # BY FIRE BE PURGED -----------------------------------------------
            getattr(self, "_apply_" + node_attr + "_shints")(new_calls)
            # -----------------------------------------------------------------
            for new_call in new_calls:
                for uv in new_call.variables():
                    hint_value = hints.get(uv.name)
                    if hint_value is not None:
                        self.log.debug("Apply runtime hint('%s', %s)",
                            uv.name, hint_value)
                    uv.resolve(hint_value)
                self._link_from_call(new_call, link_cls, rcls, collection)

    def _link_from_call(self, call, link_cls, resource_cls, collection):
        ns = RosName.resolve_ns(call.namespace, ns=self.ns, private_ns=self.pns)
        call_name = RosName(call.name, ns=ns, private_ns=self.pns)
        rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                          remaps=self.node.remaps)
        self.log.debug("_link_from_call: %s, %s, %s, %s", repr(call.namespace),
            repr(self.ns), repr(self.pns), repr(ns))
        self.log.debug("Creating %s link for %s (%s).",
            resource_cls.__name__, call_name.full, rosname.full)
        resource = collection.get(rosname.full)
        if resource is not None:
            self.log.debug("Found %s '%s' within collection.",
                resource_cls.__name__, resource.id)
        else:
            self.log.debug("No %s named '%s' was found. Creating new Resource.",
                resource_cls.__name__, rosname.full)
            resource = self._new_resource(rosname, call, resource_cls)
            collection.add(resource)
        resource.variables.extend(uv.name for uv in call.variables())
        link = link_cls.link_from_call(self.node, resource, call_name, call)
        return link

    def _new_resource(self, rosname, call, cls):
        config = self.node.configuration
        if cls is Topic:
            return Topic(config, rosname, message_type=call.type)
        elif cls is Service:
            return Service(config, rosname, message_type=call.type)
        elif cls is Parameter:
            return Parameter(config, rosname, call.type, None)
        assert False, "Unknown Resource class {}".format(cls.__name__)

    # BY FIRE BE PURGED -------------------------------------------------------

    _W_NO_EXISTS = "[Configuration '%s'][hints]: there is no %s '%s' to fix."

    _W_AMBIGUOUS = ("[Configuration '%s'][hints]: there are multiple %ss that "
                    "match '%s'.")

    def _apply_advertise_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.advertise:
            g = shint["topic"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = AdvertiseCall(g, "/", shint["msg_type"],
                    shint["queue_size"], latched=shint.get("latched", False))
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "msg_type" in shint:
                    call.type = shint["msg_type"]
                if "queue_size" in shint:
                    call.queue_size = shint["queue_size"]
                if "latched" in shint:
                    call.latched = shint["latched"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)

    def _apply_subscribe_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.subscribe:
            g = shint["topic"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = SubscribeCall(g, "/", shint["msg_type"],
                    shint["queue_size"])
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "msg_type" in shint:
                    call.type = shint["msg_type"]
                if "queue_size" in shint:
                    call.queue_size = shint["queue_size"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)

    def _apply_service_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.service:
            g = shint["service"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = AdvertiseServiceCall(g, "/", shint["srv_type"])
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "srv_type" in shint:
                    call.type = shint["srv_type"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)

    def _apply_client_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.client:
            g = shint["service"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = ServiceClientCall(g, "/", shint["srv_type"])
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "srv_type" in shint:
                    call.type = shint["srv_type"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)

    def _apply_read_param_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.read_param:
            g = shint["parameter"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = GetParamCall(g, "/", shint["param_type"],
                    default_value=shint.get("default_value"))
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "param_type" in shint:
                    call.type = shint["param_type"]
                if "default_value" in shint:
                    call.default_value = shint["default_value"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)

    def _apply_write_param_shints(self, calls):
        create = []
        config = self.node.configuration
        for shint in self.write_param:
            g = shint["parameter"]
            matches = []
            for call in calls:
                ns = RosName.resolve_ns(call.namespace,
                    ns=self.ns, private_ns=self.pns)
                rosname = RosName(call.name, ns=ns, private_ns=self.pns,
                    remaps=self.node.remaps)
                if g == rosname.full:
                    matches.append(call)
            if shint.get("create", False):
                call = SetParamCall(g, "/", shint["param_type"],
                    value=shint.get("value"))
                call.location2 = JSON_to_loc2(shint.get("traceability"))
                create.append(call)
            else:
                if not matches:
                    self.log.warning("No matches: %s", shint)
                    continue
                if len(matches) > 1:
                    self.log.warning("Many matches: %s", shint)
                    continue
                call = matches[0]
                if "original_name" in shint:
                    call.name = shint["original_name"] or "/?"
                    call.namespace = "/"
                if "param_type" in shint:
                    call.type = shint["param_type"]
                if "value" in shint:
                    call.value = shint["value"]
                if "traceability" in shint:
                    call.location2 = JSON_to_loc2(shint["traceability"])
                if "conditional" in shint:
                    v = _bool_to_conditions(not shint["conditional"])
                    call.conditions = v
        calls.extend(create)
    # -------------------------------------------------------------------------


class ConfigHints2(LoggingObject):
    def __init__(self, hints, find_node):
        self._build_hint_dict(hints, "nodes")
        self._build_hint_dict(hints, "parameters")
        self.find_node = find_node

    def apply_param_fixes(self, config):
        for name, datum in self.fix_parameters.items():
            self._fix_param(config, name, datum)

    def apply_node_fixes(self, config, futures):
        for name, datum in self.fix_nodes.items():
            self._fix_node(config, futures, name, datum)

    def create_new_params(self, config):
        for name, datum in self.add_parameters.items():
            self._add_param(config, name, datum)

    def create_new_nodes(self, config, futures):
        for name, datum in self.add_nodes.items():
            self._add_node(config, futures, name, datum)

    def _build_hint_dict(self, top_lvl_hints, key):
        to_fix = {}
        to_add = {}
        setattr(self, "fix_" + key, to_fix)
        setattr(self, "add_" + key, to_add)
        hints = top_lvl_hints.get(key)
        if hints:
            for name, datum in hints.items():
                if datum.get("create"):
                    to_add[name] = datum
                else:
                    to_fix[name] = datum

    _W_NO_EXISTS = "[Configuration '%s'][hints]: there is no %s '%s' to fix."

    _W_AMBIGUOUS = ("[Configuration '%s'][hints]: there are multiple %s that "
                    "match '%s'.")

    _W_COLLIDES = ("[Configuration '%s'][hints]: %s name '%s' already exists "
                   "for another resource.")

    _W_UNSUPPORTED = ("[Configuration '%s'][hints]: fixing %s '%s' "
                      "is not supported yet.")

    _W_EXISTS = ("[Configuration '%s'][hints]: tried to create "
                 "%s '%s', but it already exists.")

    def _fix_param(self, config, name, datum):
        params = config.parameters.get_all(name)
        if not params:
            self.log.warning(self._W_NO_EXISTS, config.name, "param", name)
            return False
        if len(params) > 1:
            self.log.warning(self._W_AMBIGUOUS, config.name, "param", name)
            return False
        param = params[0]
        if "rosname" in datum:
            new_name = datum["rosname"] or "/?"
            if config.parameters.get(new_name) is not None:
                self.log.warning(self._W_COLLIDES, config.name, "param", new_name)
                return False
            config.parameters.remove(name)
            param.rosname = RosName(new_name)
            config.parameters.add(param)
        if "param_type" in datum:
            param.type = datum["param_type"]
        if "default_value" in datum:
            param.value = datum["default_value"]
        if "traceability" in datum:
            param.location2 = JSON_to_loc2(datum["traceability"])
        if "conditional" in datum:
            v = _bool_to_conditions(not datum["conditional"])
            param.conditions = v
        return True

    def _fix_node(self, config, futures, name, datum):
        nodes = config.nodes.get_all(name)
        if not nodes:
            self.log.warning(self._W_NO_EXISTS, config.name, "node", name)
            return False
        if len(nodes) > 1:
            self.log.warning(self._W_AMBIGUOUS, config.name, "node", name)
            return False
        node = nodes[0]
        if "rosname" in datum:
            new_name = datum["rosname"] or "/?"
            if config.nodes.get(new_name) is not None:
                self.log.warning(self._W_COLLIDES, config.name, "node", new_name)
                return False
            config.nodes.remove(name)
            node.rosname = RosName(new_name)
            config.nodes.add(node)
        if "node_type" in datum:
            self.log.warning(self._W_UNSUPPORTED, config.name,
                             "node", "node_type")
        if "args" in datum:
            node.argv = datum["args"]
        if "remaps" in datum:
            node.remaps = dict(datum["remaps"])
        if "traceability" in datum:
            node.location2 = JSON_to_loc2(datum["traceability"])
        if "conditional" in datum:
            v = _bool_to_conditions(not datum["conditional"])
            node.conditions = v
        # BY FIRE BE PURGED ---------------------------------------------------
        for fnl in futures:
            if fnl.node is not node:
                continue
            fnl.advertise.extend(datum.get("publishers", ()))
            fnl.subscribe.extend(datum.get("subscribers", ()))
            fnl.service.extend(datum.get("servers", ()))
            fnl.client.extend(datum.get("clients", ()))
            fnl.read_param.extend(datum.get("getters", ()))
            fnl.write_param.extend(datum.get("setters", ()))
        # ---------------------------------------------------------------------
        return True

    def _add_param(self, config, name, datum):
        if config.parameters.get(name) is not None:
            self.log.warning(self._W_EXISTS, config.name, "param", name)
            return False
        ptype = datum["param_type"]
        value = datum.get("default_value")
        param = Parameter(config, RosName(name), ptype, value)
        param.location2 = JSON_to_loc2(datum.get("traceability"))
        config.parameters.add(param)
        return True

    def _add_node(self, config, futures, name, datum):
        if config.nodes.get(name) is not None:
            self.log.warning(self._W_EXISTS, config.name, "node", name)
            return False
        pkg, exe = datum["node_type"].split("/", 1)
        args = datum.get("args", "")
        remaps = datum.get("remaps", {})
        node = self.find_node(pkg, exe, args)
        rosname = RosName(name)
        self.log.debug("[hints] Creating NodeInstance %s for Node %s.",
                       rosname.full, node.name)
        instance = NodeInstance(self.configuration, rosname, node,
                                argv=args, remaps=remaps)
        instance.location2 = JSON_to_loc2(datum.get("traceability"))
        node.instances.append(instance)
        previous = self.configuration.nodes.add(instance)
        fnl = FutureNodeLinks(instance)
        fnl.advertise.extend(datum.get("publishers", ()))
        fnl.subscribe.extend(datum.get("subscribers", ()))
        fnl.service.extend(datum.get("servers", ()))
        fnl.client.extend(datum.get("clients", ()))
        fnl.read_param.extend(datum.get("getters", ()))
        fnl.write_param.extend(datum.get("setters", ()))
        futures.append(fnl)
        return True




#Copyright (c) 2017 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

###############################################################################
# Imports
###############################################################################

import math
import os
import re
import sys

sys.modules['_elementtree'] = None
import xml.etree.ElementTree as ET


###############################################################################
# Notes to Self
###############################################################################

# Parser extracts the XML tree. Run type conversion and substitution for things
# that are constant (e.g. <arg name="a" value="1"> and "$(arg a)").
# For unknown stuff, store a pair (type, name) and add the attribute name to an
# *unknown* list in the parsed tag element.
# Parser should report true errors (e.g. "$(arg undeclared)").

# Later, analyser picks a tag, iterates over the *unknown* and injects
# configuration context to try to resolve the remaining expressions.
# If an expression cannot be resolved inside an "if" or "unless",
# the entity is created but marked as conditional.
# If an expression cannot be resolved for some other attribute,
# a configuration error is reported.

# Draft:
#   - work on a copy, do not change original tree
#   attributes = dict(tag.attributes)
#   try:
#       for key in tag.unknown:
#           attributes[key] = resolve(attributes[key], configuration)
#       configuration.register(...)
#   except SubstitutionError as e:
#       configuration.errors.append(...)


###############################################################################
# Line and Column Numbers
###############################################################################

# courtesy of https://stackoverflow.com/a/36430270
class LineNumberingParser(ET.XMLParser):
    def _start_list(self, *args, **kwargs):
        # Here we assume the default XML parser which is expat
        # and copy its element position attributes into output Elements
        element = super(self.__class__, self)._start_list(*args, **kwargs)
        element._start_line_number = self.parser.CurrentLineNumber
        element._start_column_number = self.parser.CurrentColumnNumber + 1
        element._start_byte_index = self.parser.CurrentByteIndex
        return element

    def _end(self, *args, **kwargs):
        element = super(self.__class__, self)._end(*args, **kwargs)
        element._end_line_number = self.parser.CurrentLineNumber
        element._end_column_number = self.parser.CurrentColumnNumber + 1
        element._end_byte_index = self.parser.CurrentByteIndex
        return element


###############################################################################
# Substitution Expressions
###############################################################################

class UnresolvedValue(object):
    def __init__(self):
    # ----- parts is a list of strings and tuples, where the tuples
    #       represent the unknown bits (substitution command, value)
        self.parts = []

    def append(self, part):
        assert isinstance(part, (basestring, tuple))
        self.parts.append(part)

    @property
    def resolvable(self):
        for part in self.parts:
            if isinstance(part, tuple):
                return False
        return True

    def try_convert(self, conversion = str):
        if self.resolvable:
            return conversion("".join(self.parts))
        return self

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        ss = []
        for part in self.parts:
            if isinstance(part, tuple):
                ss.append("$({})".format(" ".join(part)))
            else:
                ss.append(part)
        return "".join(ss)


class SubstitutionError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class SubstitutionParser(object):
    PATTERN = re.compile(r"\$\(([^$()]+?)\)")
    ERROR_PATTERN = re.compile(r"\$\([^$()]*?\$[^$()]*?\)")

    COMMANDS = ("find", "env", "optenv", "dirname", "anon", "arg", "eval")

    def __init__(self, args = None, env = None, pkgs = None, anon = None,
                 dirname = None, pkg_depends = None, env_depends = None):
        self.arguments = args if not args is None else {}
        self.environment = env if not env is None else {}
        self.packages = pkgs if not pkgs is None else {}
        self.anonymous = anon if not anon is None else {}
        self.dirname = dirname
        self.pkg_depends = pkg_depends if not pkg_depends is None else set()
        self.env_depends = env_depends if not env_depends is None else set()

    def sub(self, value, conversion = str):
        """Resolve substitution arguments in the given string.
            Return a literal value if resolution is possible.
            Otherwise, return an UnresolvedValue instance.
        """
        if value.startswith("$(eval") and value.endswith(")"):
            result = UnresolvedValue()
            result.append(("eval", value[7:-1]))
            return result
        if self.ERROR_PATTERN.search(value):
            raise SubstitutionError("'$' cannot appear within expression")
        match = self.PATTERN.search(value)
        if not match:
            return self.convert_str(value, conversion)
        result = UnresolvedValue()
        rest = value
        while match:
            parts = filter(bool, map(str.strip, match.group(1).split(None, 1)))
            assert len(parts) == 1 or len(parts) == 2
            if not parts[0] in self.COMMANDS:
                raise SubstitutionError("invalid command: " + parts[0])
            prefix = rest[:match.start()]
            if prefix:
                if parts[0] == "eval":
                    raise SubstitutionError("eval must appear at the start")
                result.append(prefix)
            cmd = getattr(self, "_" + parts[0])
            result.append(cmd(parts))
            rest = rest[match.end():]
            match = self.PATTERN.search(rest)
        if rest:
            result.append(rest)
        return self.convert_unresolved(result, conversion)

    def resolve(self, value, conversion = str, strict = False):
        if not isinstance(value, UnresolvedValue):
            return value
        parts = []
        for part in value.parts:
            if isinstance(part, basestring):
                parts.append(part)
            else:
                assert isinstance(part, tuple)
                value = getattr(self, "_" + part[0])(part)
                if isinstance(value, tuple):
                    # a SubstitutionError here cannot be distinguished
                    # from one coming from getattr above
                    if not strict:
                        return None
                    raise SubstitutionError("cannot resolve: " + str(value))
                parts.append(value)
        return self.convert_str("".join(parts), conversion)

    def to_bool(self, value):
        if value is True or value == "1" or str(value).lower() == "true":
            return True
        if value is False or value == "0" or str(value).lower() == "false":
            return False
        raise SubstitutionError("invalid boolean value: " + value)

    def to_float(self, value):
        try:
            return float(value)
        except ValueError as e:
            raise SubstitutionError("invalid number value: " + value)

    def to_int(self, value):
        try:
            return int(value)
        except ValueError as e:
            raise SubstitutionError("invalid int value: " + value)

    def convert_str(self, value, conversion):
        if conversion == bool:
            return self.to_bool(value)
        if conversion == float:
            return self.to_float(value)
        if conversion == int:
            return self.to_int(value)
        return conversion(value)

    def convert_unresolved(self, value, conversion):
        if conversion == bool:
            return value.try_convert(conversion = self.to_bool)
        if conversion == float:
            return value.try_convert(conversion = self.to_float)
        if conversion == int:
            return value.try_convert(conversion = self.to_int)
        return value.try_convert(conversion = conversion)

    def _find(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("find takes exactly one argument")
        name = parts[1]
        self.pkg_depends.add(name)
        try:
            package = self.packages.get("package:" + name)
        except KeyError:
            package = None

        if package:
            if package.path:
                return package.path
            return ("find", name)
        raise SubstitutionError("unknown package: " + name)

    def _arg(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("arg takes exactly one argument")
        name = parts[1]
        if name in self.arguments:
            value = self.arguments[name]
            if value is None or isinstance(value, UnresolvedValue):
                return ("arg", name)
            return value
        raise SubstitutionError("undeclared arg: " + name)

    def _anon(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("anon takes exactly one argument")
        name = parts[1]
        if name in self.anonymous:
            return self.anonymous[name]
        value = self._anonymous_name(name)
        self.anonymous[name] = value
        return value

    def _env(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("env takes exactly one argument")
        self.env_depends.add(parts[1])
        return self.environment.get(parts[1], tuple(parts))

    def _optenv(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("optenv expects at least one argument")
        args = parts[1].split(None, 1)
        self.env_depends.add(args[0])
        default = args[1] if len(args) == 2 else ""
        return self.environment.get(parts[1], default)

    def _dirname(self, parts):
        if len(parts) > 1:
            raise SubstitutionError("dirname does not take arguments")
        if self.dirname is None:
            return ("dirname",)
        return self.dirname

    # Create a dictionary of global symbols that will be available in eval.
    # Copied from roslaunch.
    _EVAL_DICT = {
        'true': True, 'false': False,
        'True': True, 'False': False,
        '__builtins__': {k: __builtins__[k]
            for k in ('list', 'dict', 'map', 'str', 'float', 'int')},
    }

    def _eval(self, parts):
        if len(parts) != 2:
            raise SubstitutionError("eval takes exactly one argument")
        s = parts[1]
        functions = {
            "arg": self._eval_arg,
            "find": self._eval_find,
            "anon": self._eval_anon,
            "env": self._eval_env,
            "optenv": self._eval_optenv,
            "dirname": self._eval_dirname
        }
        functions.update(self._EVAL_DICT)
        # ignore values containing double underscores (for safety)
        # http://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
        if s.find('__') >= 0:
            raise SubstitutionError(
                "$(eval ...) may not contain double underscore expressions")
        try:
            r = eval(s, {}, _DictWrapper(self.arguments, functions))
            if isinstance(r, UnresolvedValue):
                return r
            return str(r)
        except SubstitutionError as e:
            raise e
        except TypeError:
            r = UnresolvedValue()
            r.append(("eval", s))
            return r

    def _anonymous_name(self, name):
        try:
            from rosgraph.names import anonymous_name
            return anonymous_name(name)
        except ImportError:
            import random, socket, sys, warnings
            warnings.warn("Could not import the 'rosgraph' package; "
                          "resorting to fallback behaviour.")
            # Behaviour copied from rosgraph.names
            name = "{}_{}_{}_{}".format(name, socket.gethostname(),
                os.getpid(), random.randint(0, sys.maxsize))
            return name.replace('.', '_').replace('-', '_').replace(':', '_')

    def _eval_arg(self, name):
        return _convert_value(self._arg(("arg", name)))

    def _eval_find(self, name):
        return self._find(("find", name))

    def _eval_anon(self, name):
        return self._anon(("anon", name))

    def _eval_env(self, name):
        return self._env(("env", name))

    def _eval_optenv(self, name, default=""):
        self.env_depends.add(name)
        return self.environment.get(name, default)

    def _eval_dirname(self):
        return ("dirname",) if self.dirname is None else self.dirname


SubstitutionParser._EVAL_DICT.update(math.__dict__)


# basically copied from roslaunch
class _DictWrapper(object):
    def __init__(self, args, functions):
        self._args = args
        self._functions = functions

    def __getitem__(self, key):
        try:
            return self._functions[key]
        except KeyError:
            if key in self._args:
                value = self._args[key]
                if value is None or isinstance(value, UnresolvedValue):
                    return ("arg", key)
                return _convert_value(value)
            raise SubstitutionError("undeclared arg: " + key)


def _convert_value(value):
    if isinstance(value, tuple):
        return value
    #attempt numeric conversion
    try:
        if '.' in value:
            return float(value)
        else:
            return int(value)
    except ValueError as e:
        pass
    #bool
    lval = value.lower()
    if lval == 'true':
        return True
    if lval == 'false':
        return False
    #string
    return value


###############################################################################
# Launch XML Parser
###############################################################################

class LaunchParserError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class ErrorTag(object):
    _ATTRIBUTES = {}
    _EMPTY_LIST = ()

    def __init__(self, text):
        self.text = text
        self.attributes = self._ATTRIBUTES
        self.children = self._EMPTY_LIST
        self.unknown = self._EMPTY_LIST

    @property
    def tag(self):
        return "error"

    def append(self, child):
        raise NotImplementedError("error nodes do not have children")


class BaseLaunchTag(object):
    CHILDREN = ()
    REQUIRED = ()
    ATTRIBUTES = {
        "if": bool,
        "unless": bool
    }

    def __init__(self, text, attributes, line, col):
        self.text = text
        self.line = line
        self.column = col
        self.attributes = attributes
        for key in self.REQUIRED:
            if not attributes.get(key):
                raise LaunchParserError("missing required attribute: " + key)
        self.children = []
        self.unknown = []
        for key, value in attributes.iteritems():
            if isinstance(value, UnresolvedValue):
                self.unknown.append(key)
        if "if" in attributes and "unless" in attributes:
            raise LaunchParserError("cannot declare both 'if' and 'unless'")
    # ----- A condition is a tuple (target, value), where target is what
    #       the condition should evaluate to ("if" = True, "unless" = False).
        if "unless" in attributes:
            self.condition = (False, attributes["unless"])
        else:
            self.condition = (True, attributes.get("if", True))

    @property
    def tag(self):
        raise NotImplementedError("subclasses must override 'tag'")

    def append(self, child):
        if child.tag in self.CHILDREN or child.tag == "error":
            self.children.append(child)
        else:
            self.children.append(ErrorTag("invalid child tag: " + child.tag))


class LaunchTag(BaseLaunchTag):
    CHILDREN = ("node", "include", "remap", "param", "rosparam",
                "group", "arg", "env", "machine", "test")
    ATTRIBUTES = {}

    @property
    def tag(self):
        return "launch"


class NodeTag(BaseLaunchTag):
    CHILDREN = ("remap", "param", "rosparam", "env")
    REQUIRED = ("pkg", "type")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "pkg": str,
        "type": str,
        "name": str,
        "args": str,
        "machine": str,
        "respawn": bool,
        "respawn_delay": float,
        "required": bool,
        "ns": str,
        "clear_params": bool,
        "output": str,
        "cwd": str,
        "launch-prefix": str
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.package = attributes["pkg"]
        self.type = attributes["type"]
        self.name = attributes.get("name")
        self.argv = attributes.get("args")
        self.machine = attributes.get("machine")
        self.respawn = attributes.get("respawn", False)
        self.respawn_delay = attributes.get("respawn_delay", 0.0)
        self.required = attributes.get("required", False)
        self.namespace = attributes.get("ns")
        self.clear_params = attributes.get("clear_params", False)
        self.output = attributes.get("output", "log")
        self.cwd = attributes.get("cwd", "ROS_HOME")
        self.prefix = attributes.get("launch-prefix")

    @property
    def tag(self):
        return "node"


class IncludeTag(BaseLaunchTag):
    CHILDREN = ("arg", "env")
    REQUIRED = ("file",)
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "file": str,
        "ns": str,
        "clear_params": bool,
        "pass_all_args": bool
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.file = attributes["file"]
        self.namespace = attributes.get("ns")
        self.clear_params = attributes.get("clear_params", False)
        self.pass_all_args = attributes.get("pass_all_args", False)

    @property
    def tag(self):
        return "include"


class RemapTag(BaseLaunchTag):
    REQUIRED = ("from", "to")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "from": str,
        "to": str
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.origin = attributes["from"]
        self.target = attributes["to"]

    @property
    def tag(self):
        return "remap"


class ParamTag(BaseLaunchTag):
    REQUIRED = ("name",)
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "name": str,
        "value": str,
        "type": str,
        "textfile": str,
        "binfile": str,
        "command": str
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.name = attributes["name"]
        self.value = attributes.get("value")
        self.type = attributes.get("type")
        self.textfile = attributes.get("textfile")
        self.binfile = attributes.get("binfile")
        self.command = attributes.get("command")
        if (self.value is None and self.textfile is None
                and self.binfile is None and self.command is None):
            raise LaunchParserError("missing required attribute: value")

    @property
    def tag(self):
        return "param"


class RosParamTag(BaseLaunchTag):
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "command": str,
        "file": str,
        "param": str,
        "ns": str,
        "subst_value": bool
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.command = attributes.get("command", "load")
        self.file = attributes.get("file")
        self.name = attributes.get("param")
        self.namespace = attributes.get("ns")
        self.substitute = attributes.get("subst_value", False)
        if self.command == "load":
            if self.file is None and not text:
                raise LaunchParserError("missing required attribute: file")
        elif self.command == "dump":
            if self.file is None:
                raise LaunchParserError("missing required attribute: file")
        elif self.command == "delete" and self.name is None:
            raise LaunchParserError("missing required attribute: name")

    @property
    def tag(self):
        return "rosparam"


class GroupTag(BaseLaunchTag):
    CHILDREN = ("node", "include", "remap", "param", "rosparam",
                "group", "arg", "env", "machine", "test")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "ns": str,
        "clear_params": bool
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.namespace = attributes.get("ns")
        self.clear_params = attributes.get("clear_params", False)

    @property
    def tag(self):
        return "group"


class ArgTag(BaseLaunchTag):
    REQUIRED = ("name",)
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "name": str,
        "value": str,
        "default": str,
        "doc": str
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.name = attributes["name"]
        self.value = attributes.get("value")
        self.default = attributes.get("default")
        self.description = attributes.get("doc")
        if not self.value is None and not self.default is None:
            raise LaunchParserError("incompatible attributes: value, default")

    @property
    def tag(self):
        return "arg"


class EnvTag(BaseLaunchTag):
    REQUIRED = ("name", "value")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "name": str,
        "value": str
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.name = attributes["name"]
        self.value = attributes["value"]

    @property
    def tag(self):
        return "env"


class MachineTag(BaseLaunchTag):
    REQUIRED = ("name", "address")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "name": str,
        "address": str,
        "env-loader": str,
        "default": bool,
        "user": str,
        "password": str,
        "timeout": float
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.name = attributes["name"]
        self.address = attributes["address"]
        self.loader = attributes.get("env-loader")
        self.default = attributes.get("default", "false")
        self.user = attributes.get("user")
        self.password = attributes.get("password")
        self.timeout = attributes.get("timeout", 10.0)

    @property
    def tag(self):
        return "machine"


class TestTag(BaseLaunchTag):
    CHILDREN = ("remap", "param", "rosparam", "env")
    REQUIRED = ("test-name", "pkg", "type")
    ATTRIBUTES = {
        "if": bool,
        "unless": bool,
        "test-name": str,
        "pkg": str,
        "type": str,
        "name": str,
        "args": str,
        "ns": str,
        "clear_params": bool,
        "cwd": str,
        "launch-prefix": str,
        "retry": int,
        "time-limit": float
    }

    def __init__(self, text, attributes, line, col):
        BaseLaunchTag.__init__(self, text, attributes, line, col)
        self.test_name = attributes["test-name"]
        self.package = attributes["pkg"]
        self.type = attributes["type"]
        self.name = attributes.get("name", self.test_name)
        self.argv = attributes.get("args")
        self.namespace = attributes.get("ns")
        self.clear_params = attributes.get("clear_params", False)
        self.cwd = attributes.get("cwd", "ROS_HOME")
        self.prefix = attributes.get("launch-prefix")
        self.retry = attributes.get("retry", 0)
        self.time_limit = attributes.get("time-limit", 60.0)

    @property
    def tag(self):
        return "test"


class LaunchParser(object):
    TAGS = {
        "launch": LaunchTag,
        "node": NodeTag,
        "include": IncludeTag,
        "remap": RemapTag,
        "param": ParamTag,
        "rosparam": RosParamTag,
        "group": GroupTag,
        "arg": ArgTag,
        "env": EnvTag,
        "machine": MachineTag,
        "test": TestTag
    }

    def __init__(self, pkgs=None, args=None, enable_defaults=False):
        self.sub_parser = None
        self.packages = pkgs if pkgs is not None else {}
        self.passed_args = args if args is not None else {}
        self.enable_defaults = enable_defaults

    def parse(self, filepath):
        if not filepath or not os.path.isfile(filepath):
            raise LaunchParserError("not a file: " + str(filepath))
        try:
            self.sub_parser = SubstitutionParser(
                pkgs=self.packages, args=dict(self.passed_args))
            xml_root = ET.parse(filepath, parser=LineNumberingParser()).getroot()
            if not xml_root.tag == "launch":
                raise LaunchParserError("invalid root tag: " + xml_root.tag)
            return self._parse_tag(xml_root)
        except ET.ParseError as e:
            raise LaunchParserError(str(e))

    def _parse_tag(self, tag):
        if not tag.tag in self.TAGS:
            return ErrorTag("unknown tag: " + tag.tag)
        cls = self.TAGS[tag.tag]
        try:
            attributes = self._attributes(tag, cls.ATTRIBUTES)
        except SubstitutionError as e:
            return ErrorTag(e.value)
        text = tag.text if tag.text else ""
        if tag.tag != "rosparam":
            text = text.strip()
        element = cls(text, attributes,
            tag._start_line_number, tag._start_column_number)
        if element.tag == "arg" and isinstance(element.name, basestring):
            if element.value is None:
                if self.enable_defaults:
                    self.sub_parser.arguments[element.name] = element.default
                else:
                    self.sub_parser.arguments[element.name] = None
            else:
                self.sub_parser.arguments[element.name] = element.value
        for child in tag:
            element.append(self._parse_tag(child))
        return element

    def _attributes(self, tag, schema):
        attributes = {}
        sub = self.sub_parser.sub # shortcut to make line below shorter
        for key, value in tag.attrib.iteritems():
            if not key in schema:
                continue # TODO raise an error vs. future compatibility
            attributes[key] = sub(value, conversion=schema[key])
        return attributes


###############################################################################
# Tests
###############################################################################

def _test_substitution():
    parser = SubstitutionParser()
    value = parser.sub("value")
    assert value == "value"
    value = parser.sub("1", int)
    assert value == 1
    value = parser.sub("1", bool)
    assert value is True
    value = parser.sub("1.0", float)
    assert value == 1.0
    value = parser.sub("$(env VAR)")
    assert isinstance(value, UnresolvedValue)
    assert len(value.parts) == 1
    assert not value.resolvable
    assert value.try_convert() is value
    value = parser.sub("$(eval 1 + 1)")
    assert isinstance(value, UnresolvedValue)
    assert len(value.parts) == 1
    value = parser.sub("value$(env NAME)$(env VAR)")
    assert isinstance(value, UnresolvedValue)
    assert len(value.parts) == 3
    assert value.parts[0] == "value"
    assert value.parts[1] == ("env", "NAME")
    assert value.parts[2] == ("env", "VAR")
    parser.arguments["test"] = "value"
    value = parser.sub("$(arg test)")
    assert value == "value"
    value = parser.sub("$$(arg test)$")
    assert value == "$value$"
    parser.environment["TEST"] = "value"
    value = parser.sub("$(env TEST)")
    assert value == "value"
    value = parser.sub("$(optenv TEST)")
    assert value == "value"
    try:
        parser.sub("$(arg $(arg name))")
        assert False
    except SubstitutionError as e:
        pass
    try:
        parser.sub("$($)")
        assert False
    except SubstitutionError as e:
        pass
    try:
        parser.sub("va$(eval 'lue')")
        assert False
    except SubstitutionError as e:
        pass
    try:
        parser.sub("value$(arg name)$(env VAR)")
        assert False
    except SubstitutionError as e:
        pass

def _test_launch():
    parser = LaunchParser()
    tree = parser.parse("minimal.launch")
    assert isinstance(tree, LaunchTag)
    assert not tree.unknown
    assert not tree.attributes
    assert not tree.text
    assert tree.condition == (True, True)
    assert len(tree.children) == 2
    assert isinstance(tree.children[0], NodeTag)
    assert isinstance(tree.children[1], NodeTag)
    node = tree.children[0]
    assert not node.text
    assert not node.unknown
    assert not node.children
    assert node.attributes["pkg"] == "fictibot_drivers"
    assert node.attributes["type"] == "fictibot_driver"
    assert node.attributes["name"] == "fictibase"
    assert node.name == "fictibase"
    assert node.package == "fictibot_drivers"
    assert node.type == "fictibot_driver"
    node = tree.children[1]
    assert not node.text
    assert not node.unknown
    assert not node.children
    assert node.attributes["pkg"] == "fictibot_controller"
    assert node.attributes["type"] == "fictibot_controller"
    assert node.attributes["name"] == "ficticontrol"
    assert node.name == "ficticontrol"
    assert node.package == "fictibot_controller"
    assert node.type == "fictibot_controller"

if __name__ == "__main__":
    _test_substitution()
    _test_launch()



#!/usr/bin/env python

#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import sys
from .haros import main

sys.exit(main(source_runner = True))




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

###############################################################################
# Imports
###############################################################################

from collections import Counter
import json
import logging
import os
import datetime
from xml.sax.saxutils import escape

from .metamodel import (
    Resource, TopicPrimitive, ServicePrimitive, ParameterPrimitive
)


###############################################################################
# Utility
###############################################################################

class LoggingObject(object):
    log = logging.getLogger(__name__)


################################################################################
# Export Manager
################################################################################

class JUnitExporter(LoggingObject):
    """
    Utility class for outputting analysis result data
    in a JUnit XML format text file.
    """
    def export_report(self, datadir, database):
        """
        Output the analysis data in a JUnit XML format text file.
        @param datadir:   [str] The folder / file system path where to store the output.
        @param database:  [.data.HarosDatabase] Database with analysis result data.
        """
        self.log.info("Exporting JUnit XML format report data.")
        if database.report == None:
            return
        report = database.report # .data.AnalysisReport
        summary_report_filename = os.path.join(
            datadir,
            database.project.name,
            "compliance",
            database.project.name+".xml"
        )
        with open(summary_report_filename, "w") as srf:
            # count how many rules have been violated
            # and how long all reports together took to create
            violated_rules = {}
            total_analysis_time = 0
            for package_analysis in report.by_package.viewvalues(): # .data.PackageAnalysis
                for violation in package_analysis.violations:
                    violated_rules[violation.rule.id] = violation.rule.name
                # ^ for violation in package_analysis.violations
                # Per-file violations:
                for file_analysis in package_analysis.file_analysis:
                    for violation in file_analysis.violations:
                        violated_rules[violation.rule.id] = violation.rule.name
                    # ^ for violation in file_analysis.violations
                # ^ for file_analysis in package_analysis.file_analysis
                total_analysis_time += report.analysis_time
            # ^ for package_analysis in report.by_package.viewvalues()
            summary_timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")
            srf.write('<?xml version="1.0" encoding="UTF-8" ?>\n')
            srf.write('<testsuites id="HAROS_%s_%s"' % (database.project, summary_timestamp))
            srf.write(' name="HAROS analysis result for %s (%s)"' % (database.project, summary_timestamp))
            srf.write(' tests="%i"' % (len(database.rules) * len(report.by_package)))
            srf.write(' failures="%i"' % len(violated_rules))
            srf.write(' time="%f">\n' % total_analysis_time)

            for package_analysis in report.by_package.viewvalues(): # .data.PackageAnalysis
                out = os.path.join(datadir,
                                   database.project.name,
                                   "compliance",
                                   "source",
                                   package_analysis.package.name + ".xml")
                try:
                    self._export_package_report(out, package_analysis, database, srf)
                except:
                    self.log.error("Failed to write JUnit XML report file: " + out)
            # ^ for package_analysis in report.by_package.viewvalues()
            srf.write('</testsuites>\n')
    # ^ def export_report(self, datadir, report)
    
    def _export_package_report(self, out, package_analysis, database, srf):
        """
        Output the analysis data for one package in a JUnit XML format text file.
        :param out:             [str] The file system path where to store the output.
        :param package_analysis [.data.PackageAnalysis] Analysis data for this package.
        :param database:        [.data.HarosDatabase] Database with analysis result data.
        :param srf              [file] Summary report file.
        """
        report = database.report # .data.AnalysisReport
        with open(out, "w") as prf:
            # count how many rules have been violated
            violated_rules = {}
            for violation in package_analysis.violations:
                violated_rules[violation.rule.id] = violation.rule.name
            # ^ for violation in package_analysis.violations
            # Per-file violations:
            for file_analysis in package_analysis.file_analysis:
                for violation in file_analysis.violations:
                    violated_rules[violation.rule.id] = violation.rule.name
                # ^ for violation in file_analysis.violations
            # ^ for file_analysis in package_analysis.file_analysis
            prf.write('<?xml version="1.0" encoding="UTF-8" ?>\n')
            prf.write('<testsuites id="HAROS_%s_%s"' % (package_analysis.package.name, report.timestamp))
            prf.write(' name="HAROS analysis result for %s (%s)"' % (package_analysis.package.name, report.timestamp))
            prf.write(' tests="%i"' % len(database.rules))
            prf.write(' failures="%i"' % len(violated_rules))
            prf.write(' time="%f">\n' % report.analysis_time)
            prf.write('  <testsuite id="HAROS.AnalysisReport.%s"' % package_analysis.package.name)
            srf.write('  <testsuite id="HAROS.AnalysisReport.%s"' % package_analysis.package.name)
            prf.write(' name="HAROS package analysis for %s"' % package_analysis.package.name)
            srf.write(' name="HAROS package analysis for %s"' % package_analysis.package.name)
            prf.write(' tests="%i"' % len(database.rules))
            srf.write(' tests="%i"' % len(database.rules))
            prf.write(' failures="%i"' % len(violated_rules))
            srf.write(' failures="%i"' % len(violated_rules))
            prf.write(' time="%f">\n' % report.analysis_time)
            srf.write(' time="%f">\n' % report.analysis_time)
            #
            # Global violations
            for violation in package_analysis.violations:
                _description = escape(violation.rule.description, {'"':'&quot;', "'":'&apos;'})
                prf.write('    <testcase id="%s"' % violation.rule.id)
                srf.write('    <testcase id="%s"' % violation.rule.id)
                prf.write(' name="%s">\n' % violation.rule.name)
                srf.write(' name="%s">\n' % violation.rule.name)
                prf.write('      <failure message="%s"' % _description)
                srf.write('      <failure message="%s"' % _description)
                prf.write(' type="%s">\n' % violation.rule.id)
                srf.write(' type="%s">\n' % violation.rule.id)
                prf.write('%s\n' % _description)
                srf.write('%s\n' % _description)
                prf.write('Category: %s\n' % violation.rule.id)
                srf.write('Category: %s\n' % violation.rule.id)
                prf.write('File: [GLOBAL]\n')
                srf.write('File: [GLOBAL]\n')
                prf.write('Line: 0\n')
                srf.write('Line: 0\n')
                prf.write('      </failure>\n')
                srf.write('      </failure>\n')
                prf.write('    </testcase>\n')
                srf.write('    </testcase>\n')
            # ^ for violation in package_analysis.violations
            # Per-file violations:
            for file_analysis in package_analysis.file_analysis:
                for violation in file_analysis.violations:
                    filename = "[UNKNOWN]"
                    line = 0
                    if violation.location != None:
                        if violation.location.file != None and violation.location.file.full_name != None:
                            filename = violation.location.file.full_name
                        if violation.location.line != None:
                            line = violation.location.line
                    _description = escape(violation.rule.description, {'"':'&quot;', "'":'&apos;'})
                    prf.write('    <testcase id="%s"' % violation.rule.id)
                    srf.write('    <testcase id="%s"' % violation.rule.id)
                    prf.write(' name="%s">\n' % violation.rule.name)
                    srf.write(' name="%s">\n' % violation.rule.name)
                    prf.write('      <failure message="%s"' % _description)
                    srf.write('      <failure message="%s"' % _description)
                    prf.write(' type="%s">\n' % violation.rule.id)
                    srf.write(' type="%s">\n' % violation.rule.id)
                    prf.write('%s\n' % _description)
                    srf.write('%s\n' % _description)
                    prf.write('Category: %s\n' % violation.rule.id)
                    srf.write('Category: %s\n' % violation.rule.id)
                    prf.write('File: %s\n' % filename)
                    srf.write('File: %s\n' % filename)
                    prf.write('Line: %i\n' % line)
                    srf.write('Line: %i\n' % line)
                    prf.write('      </failure>\n')
                    srf.write('      </failure>\n')
                    prf.write('    </testcase>\n')
                    srf.write('    </testcase>\n')
                # ^ for violation in file_analysis.violations
            # ^ for file_analysis in package_analysis.file_analysis
            prf.write('  </testsuite>\n')
            srf.write('  </testsuite>\n')
            prf.write('</testsuites>\n')
        # ^ with open(out, "w") as f
    # ^ def _write_report_file(out, package_analysis, database, srf)
# ^ class JUnitExporter


class JsonExporter(LoggingObject):
    def export_projects(self, datadir, projects, overwrite = True):
        self.log.info("Exporting project data.")
        out = os.path.join(datadir, "projects.json")
        if not overwrite and os.path.isfile(out):
            with open(out, "r") as f:
                data = json.load(f)
            for p in projects:
                is_new = True
                for i in xrange(len(data)):
                    if data[i]["id"] == p.name:
                        is_new = False
                        data[i] = p.to_JSON_object()
                        break
                if is_new:
                    data.append(p.to_JSON_object())
        else:
            data = [p.to_JSON_object() for p in projects]
        with open(out, "w") as f:
            self.log.debug("Writing to %s", out)
            json.dump(data, f, indent=2, separators=(",", ":"))

    def export_packages(self, datadir, packages):
        self.log.info("Exporting package data.")
        out = os.path.join(datadir, "packages.json")
        if isinstance(packages, dict):
            packages = packages.viewvalues()
        with open(out, "w") as f:
            self.log.debug("Writing to %s", out)
            json.dump([self._pkg_analysis_JSON(pkg) for pkg in packages], f,
                      indent=2, separators=(",", ":"))

    def export_rules(self, datadir, rules):
        self.log.info("Exporting analysis rules.")
        self._export_collection(datadir, rules, "rules.json")

    def export_metrics(self, datadir, metrics):
        self.log.info("Exporting analysis metrics.")
        self._export_collection(datadir, metrics, "metrics.json")

    def export_source_violations(self, datadir, pkg_reports):
        self.log.info("Exporting reported source rule violations.")
        if isinstance(pkg_reports, dict):
            pkg_reports = pkg_reports.viewvalues()
        for report in pkg_reports:
            out = os.path.join(datadir, report.package.name + ".json")
            data = [v.to_JSON_object() for v in report.violations]
            for fa in report.file_analysis:
                data.extend(v.to_JSON_object() for v in fa.violations)
            with open(out, "w") as f:
                self.log.debug("Writing to %s", out)
                json.dump(data, f, indent=2, separators=(",", ":"))

    def export_runtime_violations(self, datadir, config_reports):
        self.log.info("Exporting reported runtime rule violations.")
        if isinstance(config_reports, dict):
            config_reports = config_reports.viewvalues()
        for report in config_reports:
            self._export_collection(datadir, report.violations,
                                    report.configuration.name + ".json")

    def export_other_violations(self, datadir, violations):
        self.log.info("Exporting reported rule violations.")
        self._export_collection(datadir, violations, "unknown.json")

    def export_measurements(self, datadir, pkg_reports):
        self.log.info("Exporting metrics measurements.")
        if isinstance(pkg_reports, dict):
            pkg_reports = pkg_reports.viewvalues()
        for report in pkg_reports:
            out = os.path.join(datadir, report.package.name + ".json")
            data = [m.to_JSON_object() for m in report.metrics]
            for fa in report.file_analysis:
                data.extend(m.to_JSON_object() for m in fa.metrics)
            with open(out, "w") as f:
                self.log.debug("Writing to %s", out)
                json.dump(data, f, indent=2, separators=(",", ":"))

    def export_configurations(self, datadir, config_reports):
        self.log.info("Exporting launch configurations.")
        if isinstance(config_reports, dict):
            config_reports = config_reports.viewvalues()
        configs = []
        for report in config_reports:
            data = report.configuration.to_JSON_object()
            data["unresolved"] = report.configuration.get_unresolved()
            data["conditional"] = report.configuration.get_conditional()
            queries = []
            for datum in report.violations:
                if not datum.affected:
                    continue
                objects = []
                for obj in datum.affected:
                    obj_json = self._query_object_JSON(obj, report.configuration)
                    if not obj_json is None:
                        objects.append(obj_json)
                if not objects:
                    continue
                queries.append({
                    "qid": len(queries),
                    "rule": datum.rule.id,
                    "name": datum.rule.name,
                    "objects": objects
                })
            data["queries"] = queries
            configs.append(data)
        out = os.path.join(datadir, "configurations.json")
        with open(out, "w") as f:
            self.log.debug("Writing to %s", out)
            json.dump([config for config in configs], f,
                      indent=2, separators=(",", ":"))

    def export_summary(self, datadir, report, past):
        self.log.info("Exporting analysis summary.")
        out = os.path.join(datadir, "summary.json")
        data = report.to_JSON_object()
        data["history"] = {
            "timestamps": [r.timestamp for r in past],
            "lines_of_code": [r.statistics.lines_of_code for r in past],
            "comments": [r.statistics.comment_lines for r in past],
            "issues": [r.statistics.issue_count for r in past],
            "standards": [r.statistics.standard_issue_count for r in past],
            "metrics": [r.statistics.metrics_issue_count for r in past],
            "complexity": [r.statistics.avg_complexity for r in past],
            "function_length": [r.statistics.avg_function_length for r in past]
        }
        stats = report.statistics
        data["history"]["timestamps"].append(report.timestamp)
        data["history"]["lines_of_code"].append(stats.lines_of_code)
        data["history"]["comments"].append(stats.comment_lines)
        data["history"]["issues"].append(stats.issue_count)
        data["history"]["standards"].append(stats.standard_issue_count)
        data["history"]["metrics"].append(stats.metrics_issue_count)
        data["history"]["complexity"].append(stats.avg_complexity)
        data["history"]["function_length"].append(stats.avg_function_length)
        with open(out, "w") as f:
            self.log.debug("Writing to %s", out)
            json.dump(data, f, indent=2, separators=(",", ":"))

    def _export_collection(self, datadir, items, filename):
        out = os.path.join(datadir, filename)
        if isinstance(items, dict):
            items = items.viewvalues()
        with open(out, "w") as f:
            self.log.debug("Writing to %s", out)
            json.dump([item.to_JSON_object() for item in items], f,
                      indent=2, separators=(",", ":"))

    def _query_object_JSON(self, obj, config):
        if isinstance(obj, Resource) and obj.configuration == config:
             return {
                "name": obj.id,
                "uid": str(id(obj)),
                "resourceType": obj.resource_type
            }
        elif isinstance(obj, TopicPrimitive) and obj.configuration == config:
            return {
                "node": obj.node.id,
                "node_uid": str(id(obj.node)),
                "topic": obj.topic.id,
                "topic_uid": str(id(obj.topic)),
                "resourceType": "link"
            }
        elif isinstance(obj, ServicePrimitive) and obj.configuration == config:
            return {
                "node": obj.node.id,
                "node_uid": str(id(obj.node)),
                "service": obj.service.id,
                "service_uid": str(id(obj.service)),
                "resourceType": "link"
            }
        elif isinstance(obj, ParameterPrimitive) and obj.configuration == config:
            return {
                "node": obj.node.id,
                "node_uid": str(id(obj.node)),
                "param": obj.parameter.id,
                "param_uid": str(id(obj.parameter)),
                "resourceType": "link"
            }
        return None

    def _pkg_analysis_JSON(self, pkg_analysis):
        pkg = pkg_analysis.package
        data = {
            "id": pkg.name,
            "metapackage": pkg.is_metapackage,
            "version": pkg.version,
            "description": pkg.description,
            "wiki": pkg.website,
            "repository": pkg.vcs_url,
            "bugTracker": pkg.bug_url,
            "authors": [person.name for person in pkg.authors],
            "maintainers": [person.name for person in pkg.maintainers],
            "dependencies": [name for name in pkg.dependencies.packages],
            "size": "{0:.2f}".format(pkg.size / 1000.0),
            "lines": pkg.lines,
            "sloc": pkg.sloc
        }
        violations = Counter(v.rule.id for v in pkg_analysis.violations)
        violations.update(v.rule.id for fa in pkg_analysis.file_analysis
                          for v in fa.violations)
        data["analysis"] = {
            "violations": violations,
            "metrics": {m.metric.id: m.value for m in pkg_analysis.metrics}
        }
        return data
